# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class PulseOximeterClient(SensorClient):
    """
    A sensor approximating the oxygen level.
     * 
     * **Jacdac is not suitable for medical devices and should NOT be used in any kind of device to diagnose or treat any medical conditions.**
    Implements a client for the `Pulse Oximeter <https://microsoft.github.io/jacdac-docs/services/pulseoximeter>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_oxygen_value: float = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_PULSE_OXIMETER, JD_PULSE_OXIMETER_PACK_FORMATS, role)
        self.missing_oxygen_value = missing_oxygen_value

    @property
    def oxygen(self) -> Optional[float]:
        """
        The estimated oxygen level in blood., _: %
        """
        self.refresh_reading()
        return self.register(JD_PULSE_OXIMETER_REG_OXYGEN).value(self.missing_oxygen_value)

    @property
    def oxygen_error(self) -> Optional[float]:
        """
        (Optional) The estimated error on the reported sensor data., _: %
        """
        return self.register(JD_PULSE_OXIMETER_REG_OXYGEN_ERROR).value()

    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client, EventHandlerFn, UnsubscribeFn
from .constants import *
from typing import Optional, Tuple


class VerifiedTelemetryClient(Client):
    """
    A mixin service that exposes verified telemetry information for a sensor (see https://github.com/Azure/Verified-Telemetry/tree/main/PnPModel).
    Implements a client for the `Verified Telemetry <https://microsoft.github.io/jacdac-docs/services/verifiedtelemetrysensor>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_VERIFIED_TELEMETRY, JD_VERIFIED_TELEMETRY_PACK_FORMATS, role)


    @property
    def telemetry_status(self) -> Optional[VerifiedTelemetryStatus]:
        """
        Reads the telemetry working status, where ``true`` is working and ``false`` is faulty., 
        """
        return self.register(JD_VERIFIED_TELEMETRY_REG_TELEMETRY_STATUS).value()

    @property
    def telemetry_status_interval(self) -> Optional[int]:
        """
        (Optional) Specifies the interval between computing the fingerprint information., _: ms
        """
        return self.register(JD_VERIFIED_TELEMETRY_REG_TELEMETRY_STATUS_INTERVAL).value()

    @telemetry_status_interval.setter
    def telemetry_status_interval(self, value: int) -> None:
        self.register(JD_VERIFIED_TELEMETRY_REG_TELEMETRY_STATUS_INTERVAL).set_values(value)


    @property
    def fingerprint_type(self) -> Optional[VerifiedTelemetryFingerprintType]:
        """
        Type of the fingerprint., 
        """
        return self.register(JD_VERIFIED_TELEMETRY_REG_FINGERPRINT_TYPE).value()

    @property
    def fingerprint_template(self) -> Optional[Tuple[int, bytes]]:
        """
        Template Fingerprint information of a working sensor., confidence: %
        """
        return self.register(JD_VERIFIED_TELEMETRY_REG_FINGERPRINT_TEMPLATE).value()

    def on_telemetry_status_change(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        The telemetry status of the device was updated.
        """
        return self.on_event(JD_VERIFIED_TELEMETRY_EV_TELEMETRY_STATUS_CHANGE, handler)

    def on_fingerprint_template_change(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        The fingerprint template was updated
        """
        return self.on_event(JD_VERIFIED_TELEMETRY_EV_FINGERPRINT_TEMPLATE_CHANGE, handler)


    def reset_fingerprint_template(self, ) -> None:
        """
        This command will clear the template fingerprint of a sensor and collect a new template fingerprint of the attached sensor.
        """
        self.send_cmd_packed(JD_VERIFIED_TELEMETRY_CMD_RESET_FINGERPRINT_TEMPLATE, )

    def retrain_fingerprint_template(self, ) -> None:
        """
        This command will append a new template fingerprint to the `fingerprintTemplate`. Appending more fingerprints will increase the accuracy in detecting the telemetry status.
        """
        self.send_cmd_packed(JD_VERIFIED_TELEMETRY_CMD_RETRAIN_FINGERPRINT_TEMPLATE, )
    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient, EventHandlerFn, UnsubscribeFn
from .constants import *
from typing import Optional


class SwitchClient(SensorClient):
    """
    A switch, which keeps its position.
    Implements a client for the `Switch <https://microsoft.github.io/jacdac-docs/services/switch>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_active_value: bool = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_SWITCH, JD_SWITCH_PACK_FORMATS, role)
        self.missing_active_value = missing_active_value

    @property
    def active(self) -> Optional[bool]:
        """
        Indicates whether the switch is currently active (on)., 
        """
        self.refresh_reading()
        return self.register(JD_SWITCH_REG_ACTIVE).bool_value(self.missing_active_value)

    @property
    def variant(self) -> Optional[SwitchVariant]:
        """
        (Optional) Describes the type of switch used., 
        """
        return self.register(JD_SWITCH_REG_VARIANT).value()

    def on_on(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when switch goes from `off` to `on`.
        """
        return self.on_event(JD_SWITCH_EV_ON, handler)

    def on_off(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when switch goes from `on` to `off`.
        """
        return self.on_event(JD_SWITCH_EV_OFF, handler)

    



import struct
from typing import Any, List, Optional, Tuple, Union, cast

_fmts = {
    "u64": "<Q",
    "u32": "<I",
    "u16": "<H",
    "u8":  "<B",
    "i64": "<q",
    "i32": "<i",
    "i16": "<h",
    "i8":  "<b",
    "f32": "<f",
    "f64": "<d",
}


class _TokenParser:
    # c0: number
    # size: number
    # div: number
    # fp = 0
    # nfmt: NumberFormat
    # nfmt2: boolean
    # word: string
    # is_array: boolean

    def __init__(self, fmt: str):
        self.fmt = fmt
        self.fp: int = 0

    def parse(self):
        self.div: int = 1
        self.is_array: bool = False

        fmt = self.fmt
        while self.fp < len(fmt):
            endp = self.fp
            while endp < len(fmt) and not fmt[endp].isspace():
                endp += 1
            word = fmt[self.fp:endp]
            self.fp = endp + 1
            if word == "":
                continue

            dot_idx = word.index(".") if "." in word else -1
            c0 = word[0]
            # "u10.6" -> "u16", div = 1 << 6
            if (c0 == "i" or c0 == "u") and dot_idx >= 0:
                sz0 = int(word[1: dot_idx])
                sz1 = int(word[dot_idx + 1:])
                word = "%s%d" % (word[0], sz0 + sz1)
                self.div = 1 << sz1

            self.size = -1
            c1 = ""

            if word.endswith("[]"):
                word = word[0:-2]
                self.is_array = True

            if len(word) >= 2:
                c1 = word[1]
                if c1 == "[":
                    ep = word.index("]")
                    self.size = int(word[2:ep])

            self.word = word
            if word in _fmts:
                self.nfmt = _fmts[word]
                self.size = struct.calcsize(self.nfmt)
                self.c0: str = ""
            else:
                self.nfmt = None
                if c0 == "r":
                    if c1 != ":":
                        c0 = ""
                elif c0 == "s" or c0 == "b" or c0 == "x":
                    if len(word) != 1 and self.size == -1:
                        c0 = ""
                elif c0 == "z":
                    if len(word) != 1:
                        c0 = ""
                else:
                    c0 = ""
                if c0 == "":
                    raise ValueError("invalid format: %s" % word)
                self.c0 = c0

            return True
        return False


BasePackType = Union[int, float, str, bytes]
PackType = Union[BasePackType, List[BasePackType], Tuple[BasePackType]]
PackTuple = Tuple[PackType, ...]

def _jdunpack_core(buf: bytes, fmt: str, repeat: int) -> PackTuple:
    repeat_res: List[PackTuple] = []
    res: List[PackType] = []
    off: int = 0
    fp0 = 0
    parser = _TokenParser(fmt)
    if repeat and len(buf) == 0:
        return ()
    while parser.parse():
        if parser.is_array and not repeat:
            return tuple(res) + _jdunpack_core(buf[off:], fmt[fp0:], 1)

        fp0 = parser.fp
        sz = parser.size
        c0 = parser.c0
        if c0 == "z":
            endoff = off
            while (endoff < len(buf) and buf[endoff] != 0):
                endoff += 1
            sz = endoff - off
        elif sz < 0:
            sz = len(buf) - off

        if parser.nfmt is not None:
            v = struct.unpack_from(parser.nfmt, buf, off)[0]
            if (parser.div != 1):
                v /= parser.div
            res.append(v)
            off += parser.size
        else:
            subbuf = buf[off:off+sz]
            if c0 == "z" or c0 == "s":
                zerop = 0
                while (zerop < len(subbuf) and subbuf[zerop] != 0):
                    zerop += 1
                res.append(subbuf[0:zerop].decode("utf8"))
            elif c0 == "b":
                res.append(subbuf)
            elif c0 == "x":
                pass  # skip padding
            elif c0 == "r":
                return tuple(res) + _jdunpack_core(subbuf, fmt[fp0:], 2)
            else:
                assert False
            off += len(subbuf)
            if (c0 == "z"):
                off += 1

        if repeat and parser.fp >= len(fmt):
            parser.fp = 0
            if repeat == 2:
                repeat_res.append(tuple(res))
                res = []
            if (off >= len(buf)):
                break

    if repeat == 2:
        if len(res):
            repeat_res.append(tuple(res))
        return tuple(repeat_res)  # type: ignore
    else:
        return tuple(res)


def jdunpack(buf: bytes, fmt: str) -> PackTuple:
    if fmt in _fmts:
        t = struct.unpack(_fmts[fmt], buf)
        return (t[0],)
    return _jdunpack_core(buf, fmt, 0)


def _jdpack_core(trg: Optional[bytearray], fmt: str, data: List[PackType], off: int) -> int:
    idx = 0
    parser = _TokenParser(fmt)
    while parser.parse():
        c0 = parser.c0

        if c0 == "x":
            # skip padding
            off += parser.size
            continue

        if c0 == "r":
            fmt0 = fmt[parser.fp:]
            for velt in data[idx:]:
                off = _jdpack_core(trg, fmt0, cast(List[PackType], velt), off)
            idx = len(data)
            break

        if (parser.is_array):
            arr = data[idx:]
            idx = len(data)
        else:
            arr = [data[idx]]
            idx += 1

        for v in arr:
            if parser.nfmt is not None:
                if not isinstance(v, int) and not isinstance(v, float):
                    raise ValueError("expecting number, got %s" % v)
                if trg:
                    vp = int(v * parser.div)
                    struct.pack_into(parser.nfmt, trg, off, vp)
                off += parser.size
            else:
                if isinstance(v, str):
                    if (c0 == "z"):
                        buf = (v + "\u0000").encode("utf-8")
                    elif (c0 == "s"):
                        buf = v.encode("utf-8")
                    else:
                        raise ValueError("unexpected string")
                elif isinstance(v, bytes) or isinstance(v, bytearray):
                    # assume buffer
                    if (c0 == "b"):
                        buf = cast(bytes, v)
                    else:
                        raise ValueError("unexpected buffer")
                else:
                    raise ValueError("expecting string, bytes, bytearray")

                sz = parser.size
                if sz >= 0:
                    if len(buf) > sz:
                        buf = buf[0:sz]
                else:
                    sz = len(buf)

                if trg:
                    trg[off:off+len(buf)] = buf
                off += sz

    if len(data) > idx:
        raise ValueError("format too short")

    return off


def jdpack(fmt: str, *args: PackType) -> bytes:
    if fmt in _fmts:
        assert len(args) == 1
        return struct.pack(_fmts[fmt], args[0])

    data = cast(List[PackType], args)
    k = _jdpack_core(None, fmt, data, 0)
    res = bytearray(k)
    _jdpack_core(res, fmt, data, 0)
    return res


# TODO: move out to some test file?
def _jdpack_test():
    import json

    def default(o: Any):
        if isinstance(o, bytes) or isinstance(o, bytearray):
            return o.hex()
        raise TypeError("can't encode '%s'" % o)

    def stringify(o: Any):
        return json.dumps(o, default=default)

    def test_one(fmt: str, data0: List[Any], expected_payload: str = None):
        def checksame(a: Any, b: Any):
            def fail(msg: str):
                raise ValueError("jdpack test error: %s (at %s; a=%s; b=%s)" %
                                 (msg, fmt, stringify(a), stringify(b)))
            if a == b or stringify(a) == stringify(b):
                return
            fail("not the same")

        buf = jdpack(fmt, *data0)
        data1 = jdunpack(buf, fmt)
        buf_hex = buf.hex()
        print("%s->%s->%s->%s" %
              (stringify(data0), fmt, buf_hex, stringify(data1)))
        if (expected_payload is not None and expected_payload != buf_hex):
            raise ValueError("jdpack test error: payload %s, exected %s" % (
                buf_hex, expected_payload))
        checksame(data0, data1)

    test_one("i8", [-42])
    test_one("u16", [42])
    test_one("u16 u16 i16", [42, 77, -10])
    test_one("u16 z s", [42, "foo", "bar"])
    test_one("u32 z s", [42, "foo", "bar"])
    test_one("i8 z s", [42, "foo", "bar"])
    test_one("u8 z s", [42, "foo12", "bar"])
    test_one("u8 r: u8 z", [42, [17, "xy"], [18, "xx"]])
    test_one("z b", ["foo12", b'bar'])
    test_one("u16 r: u16", [42, [17], [18]])
    test_one("i8 s[9] u16 s[10] u8", [-100, "foo", 1000, "barbaz", 250])
    test_one("i8 x[4] s[9] u16 x[2] s[10] x[3] u8",
             [-100, "foo", 1000, "barbaz", 250])
    test_one("u16 u16[]", [42, 17, 18])
    test_one("u16 u16[]", [42, 18])
    test_one("u16 u16[]", [42])
    test_one("u16 z[]", [42, "foo", "bar", "bz"])
    test_one(
        "b[8] u32 u8 s",
        [b'\xa1\xb2\xc3\xd4\xe5\xf6\xa7\xb8', 0x12345678, 0x42, "barbaz"],
        expected_payload="a1b2c3d4e5f6a7b8785634124262617262617a"
    )


if __name__ == "__main__":
    _jdpack_test()



# Autogenerated constants for Soil moisture service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_SOIL_MOISTURE = const(0x1d4aa3b3)


class SoilMoistureVariant(IntEnum):
    RESISTIVE = const(0x1)
    CAPACITIVE = const(0x2)


JD_SOIL_MOISTURE_REG_MOISTURE = const(JD_REG_READING)
JD_SOIL_MOISTURE_REG_MOISTURE_ERROR = const(JD_REG_READING_ERROR)
JD_SOIL_MOISTURE_REG_VARIANT = const(JD_REG_VARIANT)
JD_SOIL_MOISTURE_PACK_FORMATS = {
    JD_SOIL_MOISTURE_REG_MOISTURE: "u0.16",
    JD_SOIL_MOISTURE_REG_MOISTURE_ERROR: "u0.16",
    JD_SOIL_MOISTURE_REG_VARIANT: "u8"
}



# Autogenerated constants for Motor service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_MOTOR = const(0x17004cd8)
JD_MOTOR_REG_SPEED = const(JD_REG_VALUE)
JD_MOTOR_REG_ENABLED = const(JD_REG_INTENSITY)
JD_MOTOR_REG_LOAD_TORQUE = const(0x180)
JD_MOTOR_REG_LOAD_ROTATION_SPEED = const(0x181)
JD_MOTOR_REG_REVERSIBLE = const(0x182)
JD_MOTOR_PACK_FORMATS = {
    JD_MOTOR_REG_SPEED: "i1.15",
    JD_MOTOR_REG_ENABLED: "u8",
    JD_MOTOR_REG_LOAD_TORQUE: "u16.16",
    JD_MOTOR_REG_LOAD_ROTATION_SPEED: "u16.16",
    JD_MOTOR_REG_REVERSIBLE: "u8"
}



# Autogenerated constants for Magnetic field level service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_MAGNETIC_FIELD_LEVEL = const(0x12fe180f)


class MagneticFieldLevelVariant(IntEnum):
    ANALOG_NS = const(0x1)
    ANALOG_N = const(0x2)
    ANALOG_S = const(0x3)
    DIGITAL_NS = const(0x4)
    DIGITAL_N = const(0x5)
    DIGITAL_S = const(0x6)


JD_MAGNETIC_FIELD_LEVEL_REG_STRENGTH = const(JD_REG_READING)
JD_MAGNETIC_FIELD_LEVEL_REG_DETECTED = const(0x181)
JD_MAGNETIC_FIELD_LEVEL_REG_VARIANT = const(JD_REG_VARIANT)
JD_MAGNETIC_FIELD_LEVEL_EV_ACTIVE = const(JD_EV_ACTIVE)
JD_MAGNETIC_FIELD_LEVEL_EV_INACTIVE = const(JD_EV_INACTIVE)
JD_MAGNETIC_FIELD_LEVEL_PACK_FORMATS = {
    JD_MAGNETIC_FIELD_LEVEL_REG_STRENGTH: "i1.15",
    JD_MAGNETIC_FIELD_LEVEL_REG_DETECTED: "u8",
    JD_MAGNETIC_FIELD_LEVEL_REG_VARIANT: "u8"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class WindDirectionClient(SensorClient):
    """
    A sensor that measures wind direction.
    Implements a client for the `Wind direction <https://microsoft.github.io/jacdac-docs/services/winddirection>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_wind_direction_value: int = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_WIND_DIRECTION, JD_WIND_DIRECTION_PACK_FORMATS, role, preferred_interval = 1000)
        self.missing_wind_direction_value = missing_wind_direction_value

    @property
    def wind_direction(self) -> Optional[int]:
        """
        The direction of the wind., _: °
        """
        self.refresh_reading()
        return self.register(JD_WIND_DIRECTION_REG_WIND_DIRECTION).value(self.missing_wind_direction_value)

    @property
    def wind_direction_error(self) -> Optional[int]:
        """
        (Optional) Error on the wind direction reading, _: °
        """
        return self.register(JD_WIND_DIRECTION_REG_WIND_DIRECTION_ERROR).value()

    



# Autogenerated constants for Acidity service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_ACIDITY = const(0x1e9778c5)
JD_ACIDITY_REG_ACIDITY = const(JD_REG_READING)
JD_ACIDITY_REG_ACIDITY_ERROR = const(JD_REG_READING_ERROR)
JD_ACIDITY_REG_MIN_ACIDITY = const(JD_REG_MIN_READING)
JD_ACIDITY_REG_MAX_HUMIDITY = const(JD_REG_MAX_READING)
JD_ACIDITY_PACK_FORMATS = {
    JD_ACIDITY_REG_ACIDITY: "u4.12",
    JD_ACIDITY_REG_ACIDITY_ERROR: "u4.12",
    JD_ACIDITY_REG_MIN_ACIDITY: "u4.12",
    JD_ACIDITY_REG_MAX_HUMIDITY: "u4.12"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class DistanceClient(SensorClient):
    """
    A sensor that determines the distance of an object without any physical contact involved.
    Implements a client for the `Distance <https://microsoft.github.io/jacdac-docs/services/distance>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_distance_value: float = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_DISTANCE, JD_DISTANCE_PACK_FORMATS, role)
        self.missing_distance_value = missing_distance_value

    @property
    def distance(self) -> Optional[float]:
        """
        Current distance from the object, _: m
        """
        self.refresh_reading()
        return self.register(JD_DISTANCE_REG_DISTANCE).value(self.missing_distance_value)

    @property
    def distance_error(self) -> Optional[float]:
        """
        (Optional) Absolute error on the reading value., _: m
        """
        return self.register(JD_DISTANCE_REG_DISTANCE_ERROR).value()

    @property
    def min_range(self) -> Optional[float]:
        """
        (Optional) Minimum measurable distance, _: m
        """
        return self.register(JD_DISTANCE_REG_MIN_RANGE).value()

    @property
    def max_range(self) -> Optional[float]:
        """
        (Optional) Maximum measurable distance, _: m
        """
        return self.register(JD_DISTANCE_REG_MAX_RANGE).value()

    @property
    def variant(self) -> Optional[DistanceVariant]:
        """
        (Optional) Determines the type of sensor used., 
        """
        return self.register(JD_DISTANCE_REG_VARIANT).value()

    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *
from typing import Optional


class PowerSupplyClient(Client):
    """
    A power supply with a fixed or variable voltage range
    Implements a client for the `Power supply <https://microsoft.github.io/jacdac-docs/services/powersupply>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_POWER_SUPPLY, JD_POWER_SUPPLY_PACK_FORMATS, role)


    @property
    def enabled(self) -> Optional[bool]:
        """
        Turns the power supply on with `true`, off with `false`., 
        """
        return self.register(JD_POWER_SUPPLY_REG_ENABLED).bool_value()

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self.register(JD_POWER_SUPPLY_REG_ENABLED).set_values(value)


    @property
    def output_voltage(self) -> Optional[float]:
        """
        The current output voltage of the power supply. Values provided must be in the range `minimum_voltage` to `maximum_voltage`, _: V
        """
        return self.register(JD_POWER_SUPPLY_REG_OUTPUT_VOLTAGE).value()

    @output_voltage.setter
    def output_voltage(self, value: float) -> None:
        self.enabled = True
        self.register(JD_POWER_SUPPLY_REG_OUTPUT_VOLTAGE).set_values(value)


    @property
    def minimum_voltage(self) -> Optional[float]:
        """
        The minimum output voltage of the power supply. For fixed power supplies, `minimum_voltage` should be equal to `maximum_voltage`., _: V
        """
        return self.register(JD_POWER_SUPPLY_REG_MINIMUM_VOLTAGE).value()

    @property
    def maximum_voltage(self) -> Optional[float]:
        """
        The maximum output voltage of the power supply. For fixed power supplies, `minimum_voltage` should be equal to `maximum_voltage`., _: V
        """
        return self.register(JD_POWER_SUPPLY_REG_MAXIMUM_VOLTAGE).value()

    



# Autogenerated constants for Buzzer service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_BUZZER = const(0x1b57b1d7)
JD_BUZZER_REG_VOLUME = const(JD_REG_INTENSITY)
JD_BUZZER_CMD_PLAY_TONE = const(0x80)
JD_BUZZER_CMD_PLAY_NOTE = const(0x81)
JD_BUZZER_PACK_FORMATS = {
    JD_BUZZER_REG_VOLUME: "u0.8",
    JD_BUZZER_CMD_PLAY_TONE: "u16 u16 u16",
    JD_BUZZER_CMD_PLAY_NOTE: "u16 u0.16 u16"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *
from typing import Optional


class DotMatrixClient(Client):
    """
    A rectangular dot matrix display, made of monochrome LEDs or Braille pins.
    Implements a client for the `Dot Matrix <https://microsoft.github.io/jacdac-docs/services/dotmatrix>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_DOT_MATRIX, JD_DOT_MATRIX_PACK_FORMATS, role)


    @property
    def dots(self) -> Optional[bytes]:
        """
        The state of the screen where dot on/off state is
        stored as a bit, column by column. The column should be byte aligned.
        
        For example, if the display has no more than 8 rows in each column, then each byte contains bits corresponding
        to a single column. Least-significant bit is on top.
        If display has 10 rows, then each column is represented by two bytes.
        The top-most 8 rows sit in the first byte (with the least significant bit being on top),
        and the remainign 2 row sit in the second byte.
        
        The following C expression can be used to check if a given `column, row` coordinate is set:
        `dots[column * column_size + (row >> 3)] & (1 << (row & 7))`, where
        `column_size` is `(number_of_rows + 7) >> 3` (note that if number of rows is 8 or less then `column_size` is `1`),
        and `dots` is of `uint8_t*` type.
        
        The size of this register is `number_of_columns * column_size` bytes., 
        """
        return self.register(JD_DOT_MATRIX_REG_DOTS).value()

    @dots.setter
    def dots(self, value: bytes) -> None:
        self.register(JD_DOT_MATRIX_REG_DOTS).set_values(value)


    @property
    def brightness(self) -> Optional[float]:
        """
        (Optional) Reads the general brightness of the display, brightness for LEDs. `0` when the screen is off., _: /
        """
        return self.register(JD_DOT_MATRIX_REG_BRIGHTNESS).float_value(100)

    @brightness.setter
    def brightness(self, value: float) -> None:
        self.register(JD_DOT_MATRIX_REG_BRIGHTNESS).set_values(value / 100)


    @property
    def rows(self) -> Optional[int]:
        """
        Number of rows on the screen, _: #
        """
        return self.register(JD_DOT_MATRIX_REG_ROWS).value()

    @property
    def columns(self) -> Optional[int]:
        """
        Number of columns on the screen, _: #
        """
        return self.register(JD_DOT_MATRIX_REG_COLUMNS).value()

    @property
    def variant(self) -> Optional[DotMatrixVariant]:
        """
        (Optional) Describes the type of matrix used., 
        """
        return self.register(JD_DOT_MATRIX_REG_VARIANT).value()

    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *
from typing import Optional


class ModelRunnerClient(Client):
    """
    Runs machine learning models.
     * 
     * Only models with a single input tensor and a single output tensor are supported at the moment.
     * Input is provided by Sensor Aggregator service on the same device.
     * Multiple instances of this service may be present, if more than one model format is supported by a device.
    Implements a client for the `Model Runner <https://microsoft.github.io/jacdac-docs/services/modelrunner>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_MODEL_RUNNER, JD_MODEL_RUNNER_PACK_FORMATS, role)


    @property
    def auto_invoke_every(self) -> Optional[int]:
        """
        When register contains `N > 0`, run the model automatically every time new `N` samples are collected.
        Model may be run less often if it takes longer to run than `N * sampling_interval`.
        The `outputs` register will stream its value after each run.
        This register is not stored in flash., 
        """
        return self.register(JD_MODEL_RUNNER_REG_AUTO_INVOKE_EVERY).value()

    @auto_invoke_every.setter
    def auto_invoke_every(self, value: int) -> None:
        self.register(JD_MODEL_RUNNER_REG_AUTO_INVOKE_EVERY).set_values(value)


    @property
    def last_run_time(self) -> Optional[int]:
        """
        The time consumed in last model execution., _: us
        """
        return self.register(JD_MODEL_RUNNER_REG_LAST_RUN_TIME).value()

    @property
    def allocated_arena_size(self) -> Optional[int]:
        """
        Number of RAM bytes allocated for model execution., _: B
        """
        return self.register(JD_MODEL_RUNNER_REG_ALLOCATED_ARENA_SIZE).value()

    @property
    def model_size(self) -> Optional[int]:
        """
        The size of the model in bytes., _: B
        """
        return self.register(JD_MODEL_RUNNER_REG_MODEL_SIZE).value()

    @property
    def last_error(self) -> Optional[str]:
        """
        Textual description of last error when running or loading model (if any)., 
        """
        return self.register(JD_MODEL_RUNNER_REG_LAST_ERROR).value()

    @property
    def format(self) -> Optional[ModelRunnerModelFormat]:
        """
        The type of ML models supported by this service.
        `TFLite` is flatbuffer `.tflite` file.
        `ML4F` is compiled machine code model for Cortex-M4F.
        The format is typically present as first or second little endian word of model file., 
        """
        return self.register(JD_MODEL_RUNNER_REG_FORMAT).value()

    @property
    def format_version(self) -> Optional[int]:
        """
        A version number for the format., 
        """
        return self.register(JD_MODEL_RUNNER_REG_FORMAT_VERSION).value()

    @property
    def parallel(self) -> Optional[bool]:
        """
        (Optional) If present and true this service can run models independently of other
        instances of this service on the device., 
        """
        return self.register(JD_MODEL_RUNNER_REG_PARALLEL).bool_value()


    def set_model(self, model_size: int) -> None:
        """
        Open pipe for streaming in the model. The size of the model has to be declared upfront.
        The model is streamed over regular pipe data packets.
        The format supported by this instance of the service is specified in `format` register.
        When the pipe is closed, the model is written all into flash, and the device running the service may reset.
        """
        self.send_cmd_packed(JD_MODEL_RUNNER_CMD_SET_MODEL, model_size)
    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *



class HidKeyboardClient(Client):
    """
    Control a HID keyboard.
     * 
     * The codes for the key (selectors) is defined in the [HID Keyboard
     * specification](https://usb.org/sites/default/files/hut1_21.pdf), chapter 10 Keyboard/Keypad Page, page 81.
     * Modifiers are in page 87.
     * 
     * The device keeps tracks of the key state and is able to clear it all with the clear command.
    Implements a client for the `HID Keyboard <https://microsoft.github.io/jacdac-docs/services/hidkeyboard>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_HID_KEYBOARD, JD_HID_KEYBOARD_PACK_FORMATS, role)



    def clear(self, ) -> None:
        """
        Clears all pressed keys.
        """
        self.send_cmd_packed(JD_HID_KEYBOARD_CMD_CLEAR, )
    



# Autogenerated constants for Common registers and commands service
from enum import IntEnum
from jacdac.constants import *
JD_ANNOUNCE_INTERVAL = const(0x1f4)



class SystemReadingThreshold(IntEnum):
    NEUTRAL = const(0x1)
    INACTIVE = const(0x2)
    ACTIVE = const(0x3)


class SystemStatusCodes(IntEnum):
    READY = const(0x0)
    INITIALIZING = const(0x1)
    CALIBRATING = const(0x2)
    SLEEPING = const(0x3)
    WAITING_FOR_INPUT = const(0x4)
    CALIBRATION_NEEDED = const(0x64)


JD_CMD_ANNOUNCE = const(0x0)
JD_CMD_GET_REGISTER = const(0x1000)
JD_CMD_SET_REGISTER = const(0x2000)
JD_CMD_CALIBRATE = const(0x2)
JD_CMD_COMMAND_NOT_IMPLEMENTED = const(0x3)
JD_REG_INTENSITY = const(0x1)
JD_REG_VALUE = const(0x2)
JD_REG_MIN_VALUE = const(0x110)
JD_REG_MAX_VALUE = const(0x111)
JD_REG_MAX_POWER = const(0x7)
JD_REG_STREAMING_SAMPLES = const(0x3)
JD_REG_STREAMING_INTERVAL = const(0x4)
JD_REG_READING = const(0x101)
JD_REG_READING_RANGE = const(0x8)
JD_REG_SUPPORTED_RANGES = const(0x10a)
JD_REG_MIN_READING = const(0x104)
JD_REG_MAX_READING = const(0x105)
JD_REG_READING_ERROR = const(0x106)
JD_REG_READING_RESOLUTION = const(0x108)
JD_REG_INACTIVE_THRESHOLD = const(0x5)
JD_REG_ACTIVE_THRESHOLD = const(0x6)
JD_REG_STREAMING_PREFERRED_INTERVAL = const(0x102)
JD_REG_VARIANT = const(0x107)
JD_REG_STATUS_CODE = const(0x103)
JD_REG_INSTANCE_NAME = const(0x109)
JD_EV_ACTIVE = const(0x1)
JD_EV_INACTIVE = const(0x2)
JD_EV_CHANGE = const(0x3)
JD_EV_STATUS_CODE_CHANGED = const(0x4)
JD_EV_NEUTRAL = const(0x7)
JD_PACK_FORMATS = {
    JD_CMD_COMMAND_NOT_IMPLEMENTED: "u16 u16",
    JD_REG_INTENSITY: "u32",
    JD_REG_VALUE: "i32",
    JD_REG_MIN_VALUE: "i32",
    JD_REG_MAX_VALUE: "i32",
    JD_REG_MAX_POWER: "u16",
    JD_REG_STREAMING_SAMPLES: "u8",
    JD_REG_STREAMING_INTERVAL: "u32",
    JD_REG_READING: "i32",
    JD_REG_READING_RANGE: "u32",
    JD_REG_SUPPORTED_RANGES: "r: u32",
    JD_REG_MIN_READING: "i32",
    JD_REG_MAX_READING: "i32",
    JD_REG_READING_ERROR: "u32",
    JD_REG_READING_RESOLUTION: "u32",
    JD_REG_INACTIVE_THRESHOLD: "i32",
    JD_REG_ACTIVE_THRESHOLD: "i32",
    JD_REG_STREAMING_PREFERRED_INTERVAL: "u32",
    JD_REG_VARIANT: "u32",
    JD_REG_STATUS_CODE: "u16 u16",
    JD_REG_INSTANCE_NAME: "s",
    JD_EV_STATUS_CODE_CHANGED: "u16 u16"
}



# Autogenerated constants for Wind direction service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_WIND_DIRECTION = const(0x186be92b)
JD_WIND_DIRECTION_REG_WIND_DIRECTION = const(JD_REG_READING)
JD_WIND_DIRECTION_REG_WIND_DIRECTION_ERROR = const(JD_REG_READING_ERROR)
JD_WIND_DIRECTION_PACK_FORMATS = {
    JD_WIND_DIRECTION_REG_WIND_DIRECTION: "u16",
    JD_WIND_DIRECTION_REG_WIND_DIRECTION_ERROR: "u16"
}



# Autogenerated constants for DC Current Measurement service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_D_CCURRENT_MEASUREMENT = const(0x1912c8ae)
JD_D_CCURRENT_MEASUREMENT_REG_MEASUREMENT_NAME = const(0x182)
JD_D_CCURRENT_MEASUREMENT_REG_MEASUREMENT = const(JD_REG_READING)
JD_D_CCURRENT_MEASUREMENT_PACK_FORMATS = {
    JD_D_CCURRENT_MEASUREMENT_REG_MEASUREMENT_NAME: "s",
    JD_D_CCURRENT_MEASUREMENT_REG_MEASUREMENT: "f64"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient, EventHandlerFn, UnsubscribeFn
from .constants import *
from typing import Optional


class PowerClient(SensorClient):
    """
    A power-provider service.
    Implements a client for the `Power <https://microsoft.github.io/jacdac-docs/services/power>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_current_draw_value: int = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_POWER, JD_POWER_PACK_FORMATS, role)
        self.missing_current_draw_value = missing_current_draw_value

    @property
    def allowed(self) -> Optional[bool]:
        """
        Can be used to completely disable the service.
        When allowed, the service may still not be providing power, see 
        `power_status` for the actual current state., 
        """
        return self.register(JD_POWER_REG_ALLOWED).bool_value()

    @allowed.setter
    def allowed(self, value: bool) -> None:
        self.register(JD_POWER_REG_ALLOWED).set_values(value)


    @property
    def max_power(self) -> Optional[int]:
        """
        (Optional) Limit the power provided by the service. The actual maximum limit will depend on hardware.
        This field may be read-only in some implementations - you should read it back after setting., _: mA
        """
        return self.register(JD_POWER_REG_MAX_POWER).value()

    @max_power.setter
    def max_power(self, value: int) -> None:
        self.register(JD_POWER_REG_MAX_POWER).set_values(value)


    @property
    def power_status(self) -> Optional[PowerPowerStatus]:
        """
        Indicates whether the power provider is currently providing power (`Powering` state), and if not, why not.
        `Overprovision` means there was another power provider, and we stopped not to overprovision the bus., 
        """
        return self.register(JD_POWER_REG_POWER_STATUS).value()

    @property
    def current_draw(self) -> Optional[int]:
        """
        (Optional) Present current draw from the bus., _: mA
        """
        self.refresh_reading()
        return self.register(JD_POWER_REG_CURRENT_DRAW).value(self.missing_current_draw_value)

    @property
    def battery_voltage(self) -> Optional[int]:
        """
        (Optional) Voltage on input., _: mV
        """
        return self.register(JD_POWER_REG_BATTERY_VOLTAGE).value()

    @property
    def battery_charge(self) -> Optional[float]:
        """
        (Optional) Fraction of charge in the battery., _: /
        """
        return self.register(JD_POWER_REG_BATTERY_CHARGE).float_value(100)

    @property
    def battery_capacity(self) -> Optional[int]:
        """
        (Optional) Energy that can be delivered to the bus when battery is fully charged.
        This excludes conversion overheads if any., _: mWh
        """
        return self.register(JD_POWER_REG_BATTERY_CAPACITY).value()

    @property
    def keep_on_pulse_duration(self) -> Optional[int]:
        """
        (Optional) Many USB power packs need current to be drawn from time to time to prevent shutdown.
        This regulates how often and for how long such current is drawn.
        Typically a 1/8W 22 ohm resistor is used as load. This limits the duty cycle to 10%., _: ms
        """
        return self.register(JD_POWER_REG_KEEP_ON_PULSE_DURATION).value()

    @keep_on_pulse_duration.setter
    def keep_on_pulse_duration(self, value: int) -> None:
        self.register(JD_POWER_REG_KEEP_ON_PULSE_DURATION).set_values(value)


    @property
    def keep_on_pulse_period(self) -> Optional[int]:
        """
        (Optional) Many USB power packs need current to be drawn from time to time to prevent shutdown.
        This regulates how often and for how long such current is drawn.
        Typically a 1/8W 22 ohm resistor is used as load. This limits the duty cycle to 10%., _: ms
        """
        return self.register(JD_POWER_REG_KEEP_ON_PULSE_PERIOD).value()

    @keep_on_pulse_period.setter
    def keep_on_pulse_period(self, value: int) -> None:
        self.register(JD_POWER_REG_KEEP_ON_PULSE_PERIOD).set_values(value)


    def on_power_status_changed(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted whenever `power_status` changes.
        """
        return self.on_event(JD_POWER_EV_POWER_STATUS_CHANGED, handler)


    def shutdown(self, ) -> None:
        """
        Sent by the power service periodically, as broadcast.
        """
        self.send_cmd_packed(JD_POWER_CMD_SHUTDOWN, )
    



# Autogenerated constants for Air Pressure service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_AIR_PRESSURE = const(0x1e117cea)
JD_AIR_PRESSURE_REG_PRESSURE = const(JD_REG_READING)
JD_AIR_PRESSURE_REG_PRESSURE_ERROR = const(JD_REG_READING_ERROR)
JD_AIR_PRESSURE_PACK_FORMATS = {
    JD_AIR_PRESSURE_REG_PRESSURE: "u22.10",
    JD_AIR_PRESSURE_REG_PRESSURE_ERROR: "u22.10"
}



# Autogenerated constants for Rover service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_ROVER = const(0x19f4d06b)
JD_ROVER_REG_KINEMATICS = const(JD_REG_READING)
JD_ROVER_PACK_FORMATS = {
    JD_ROVER_REG_KINEMATICS: "i16.16 i16.16 i16.16 i16.16 i16.16"
}



# Autogenerated constants for Wind speed service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_WIND_SPEED = const(0x1b591bbf)
JD_WIND_SPEED_REG_WIND_SPEED = const(JD_REG_READING)
JD_WIND_SPEED_REG_WIND_SPEED_ERROR = const(JD_REG_READING_ERROR)
JD_WIND_SPEED_REG_MAX_WIND_SPEED = const(JD_REG_MAX_READING)
JD_WIND_SPEED_PACK_FORMATS = {
    JD_WIND_SPEED_REG_WIND_SPEED: "u16.16",
    JD_WIND_SPEED_REG_WIND_SPEED_ERROR: "u16.16",
    JD_WIND_SPEED_REG_MAX_WIND_SPEED: "u16.16"
}



# Autogenerated constants for Humidity service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_HUMIDITY = const(0x16c810b8)
JD_HUMIDITY_REG_HUMIDITY = const(JD_REG_READING)
JD_HUMIDITY_REG_HUMIDITY_ERROR = const(JD_REG_READING_ERROR)
JD_HUMIDITY_REG_MIN_HUMIDITY = const(JD_REG_MIN_READING)
JD_HUMIDITY_REG_MAX_HUMIDITY = const(JD_REG_MAX_READING)
JD_HUMIDITY_PACK_FORMATS = {
    JD_HUMIDITY_REG_HUMIDITY: "u22.10",
    JD_HUMIDITY_REG_HUMIDITY_ERROR: "u22.10",
    JD_HUMIDITY_REG_MIN_HUMIDITY: "u22.10",
    JD_HUMIDITY_REG_MAX_HUMIDITY: "u22.10"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *
from typing import Optional


class MotorClient(Client):
    """
    A DC motor.
    Implements a client for the `Motor <https://microsoft.github.io/jacdac-docs/services/motor>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_MOTOR, JD_MOTOR_PACK_FORMATS, role)


    @property
    def speed(self) -> Optional[float]:
        """
        Relative speed of the motor. Use positive/negative values to run the motor forwards and backwards.
        Positive is recommended to be clockwise rotation and negative counterclockwise. A speed of ``0`` 
        while ``enabled`` acts as brake., _: /
        """
        return self.register(JD_MOTOR_REG_SPEED).float_value(100)

    @speed.setter
    def speed(self, value: float) -> None:
        self.enabled = True
        self.register(JD_MOTOR_REG_SPEED).set_values(value / 100)


    @property
    def enabled(self) -> Optional[bool]:
        """
        Turn the power to the motor on/off., 
        """
        return self.register(JD_MOTOR_REG_ENABLED).bool_value()

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self.register(JD_MOTOR_REG_ENABLED).set_values(value)


    @property
    def load_torque(self) -> Optional[float]:
        """
        (Optional) Torque required to produce the rated power of an electrical motor at load speed., _: kg/cm
        """
        return self.register(JD_MOTOR_REG_LOAD_TORQUE).value()

    @property
    def load_rotation_speed(self) -> Optional[float]:
        """
        (Optional) Revolutions per minute of the motor under full load., _: rpm
        """
        return self.register(JD_MOTOR_REG_LOAD_ROTATION_SPEED).value()

    @property
    def reversible(self) -> Optional[bool]:
        """
        (Optional) Indicates if the motor can run backwards., 
        """
        return self.register(JD_MOTOR_REG_REVERSIBLE).bool_value()

    



from io import TextIOWrapper
from typing import Any, List, Optional, Tuple
from .constants import *
from json import load, dump
from os import path, makedirs
from base64 import b64decode, b64encode


def value_to_json(value: bytearray) -> str:
    return b64encode(value).hex()


def json_to_value(text: str) -> bytearray:
    return bytearray(b64decode(bytearray.fromhex(text)))


class SettingsFile:
    def __init__(self, file_name: str) -> None:
        self.file_name = file_name
        if not path.exists(path.dirname(self.file_name)):
            makedirs(path.dirname(self.file_name))

    def is_secret(self, key: str) -> bool:
        return key.startswith("$")

    def _read_settings(self, f: TextIOWrapper) -> Any:
        try:
            return load(f)
        except:
            return {}

    def read(self, key: str, secrets: Optional[bool] = False) -> bytearray:
        if key is None or not path.exists(self.file_name):
            return bytearray(0)
        value = bytearray(0)
        with open(self.file_name, "r+") as f:
            settings = self._read_settings(f)
            if key in settings:
                secret = self.is_secret(key)
                if not secrets and secret:  # don't return value
                    value = bytearray(1)
                    value[0] = 0
                else:
                    value = json_to_value(settings[key])
        return value

    def delete(self, key: str) -> None:
        if key is None or not path.exists(self.file_name):
            return

        with open(self.file_name, "r+") as f:
            settings = self._read_settings(f)
            del settings[key]
            dump(settings, f)

    def write(self, key: str, value: bytearray) -> None:
        settings = {}
        if path.exists(self.file_name):
            with open(self.file_name, "r") as f:
                settings = self._read_settings(f)
        with open(self.file_name, "w+") as f:
            settings[key] = value_to_json(value)
            dump(settings, f, skipkeys=True, sort_keys=True, indent=2)

    def clear(self):
        if not path.exists(self.file_name):
            return
        with open(self.file_name, "wt") as f:
            f.write("{}")

    def list(self, secrets: bool = False) -> List[Tuple[str, bytearray]]:
        if not path.exists(self.file_name):
            return []
        with open(self.file_name, "rt") as f:
            settings = self._read_settings(f)
            resp: List[Tuple[str, bytearray]] = []
            keys = settings.keys()
            for key in keys:
                value = bytearray(0)
                secret = self.is_secret(key)
                if not secrets and secret:  # don't return value
                    value = bytearray(1)
                    value[0] = 0
                else:
                    value = json_to_value(settings[key])
            return resp

    def list_keys(self) -> List[str]:
        if not path.exists(self.file_name):
            return []
        with open(self.file_name, "rt") as f:
            settings = self._read_settings(f)
            keys = settings.keys()
            return keys



# Autogenerated constants for Magnetometer service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_MAGNETOMETER = const(0x13029088)
JD_MAGNETOMETER_REG_FORCES = const(JD_REG_READING)
JD_MAGNETOMETER_REG_FORCES_ERROR = const(JD_REG_READING_ERROR)
JD_MAGNETOMETER_CMD_CALIBRATE = const(JD_CMD_CALIBRATE)
JD_MAGNETOMETER_PACK_FORMATS = {
    JD_MAGNETOMETER_REG_FORCES: "i32 i32 i32",
    JD_MAGNETOMETER_REG_FORCES_ERROR: "i32"
}



# Autogenerated constants for Light level service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_LIGHT_LEVEL = const(0x17dc9a1c)


class LightLevelVariant(IntEnum):
    PHOTO_RESISTOR = const(0x1)
    REVERSE_BIASED_LED = const(0x2)


JD_LIGHT_LEVEL_REG_LIGHT_LEVEL = const(JD_REG_READING)
JD_LIGHT_LEVEL_REG_LIGHT_LEVEL_ERROR = const(JD_REG_READING_ERROR)
JD_LIGHT_LEVEL_REG_VARIANT = const(JD_REG_VARIANT)
JD_LIGHT_LEVEL_PACK_FORMATS = {
    JD_LIGHT_LEVEL_REG_LIGHT_LEVEL: "u0.16",
    JD_LIGHT_LEVEL_REG_LIGHT_LEVEL_ERROR: "u0.16",
    JD_LIGHT_LEVEL_REG_VARIANT: "u8"
}



# Autogenerated constants for Verified Telemetry service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_VERIFIED_TELEMETRY = const(0x2194841f)


class VerifiedTelemetryStatus(IntEnum):
    UNKNOWN = const(0x0)
    WORKING = const(0x1)
    FAULTY = const(0x2)


class VerifiedTelemetryFingerprintType(IntEnum):
    FALL_CURVE = const(0x1)
    CURRENT_SENSE = const(0x2)
    CUSTOM = const(0x3)


JD_VERIFIED_TELEMETRY_REG_TELEMETRY_STATUS = const(0x180)
JD_VERIFIED_TELEMETRY_REG_TELEMETRY_STATUS_INTERVAL = const(0x80)
JD_VERIFIED_TELEMETRY_REG_FINGERPRINT_TYPE = const(0x181)
JD_VERIFIED_TELEMETRY_REG_FINGERPRINT_TEMPLATE = const(0x182)
JD_VERIFIED_TELEMETRY_CMD_RESET_FINGERPRINT_TEMPLATE = const(0x80)
JD_VERIFIED_TELEMETRY_CMD_RETRAIN_FINGERPRINT_TEMPLATE = const(0x81)
JD_VERIFIED_TELEMETRY_EV_TELEMETRY_STATUS_CHANGE = const(JD_EV_CHANGE)
JD_VERIFIED_TELEMETRY_EV_FINGERPRINT_TEMPLATE_CHANGE = const(0x80)
JD_VERIFIED_TELEMETRY_PACK_FORMATS = {
    JD_VERIFIED_TELEMETRY_REG_TELEMETRY_STATUS: "u8",
    JD_VERIFIED_TELEMETRY_REG_TELEMETRY_STATUS_INTERVAL: "u32",
    JD_VERIFIED_TELEMETRY_REG_FINGERPRINT_TYPE: "u8",
    JD_VERIFIED_TELEMETRY_REG_FINGERPRINT_TEMPLATE: "u16 b",
    JD_VERIFIED_TELEMETRY_EV_TELEMETRY_STATUS_CHANGE: "u8"
}



# Autogenerated constants for Sound Recorder with Playback service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_SOUND_RECORDER_WITH_PLAYBACK = const(0x1b72bf50)


class SoundRecorderWithPlaybackStatus(IntEnum):
    IDLE = const(0x0)
    RECORDING = const(0x1)
    PLAYING = const(0x2)


JD_SOUND_RECORDER_WITH_PLAYBACK_CMD_PLAY = const(0x80)
JD_SOUND_RECORDER_WITH_PLAYBACK_CMD_RECORD = const(0x81)
JD_SOUND_RECORDER_WITH_PLAYBACK_CMD_CANCEL = const(0x82)
JD_SOUND_RECORDER_WITH_PLAYBACK_REG_STATUS = const(0x180)
JD_SOUND_RECORDER_WITH_PLAYBACK_REG_TIME = const(0x181)
JD_SOUND_RECORDER_WITH_PLAYBACK_REG_VOLUME = const(JD_REG_INTENSITY)
JD_SOUND_RECORDER_WITH_PLAYBACK_PACK_FORMATS = {
    JD_SOUND_RECORDER_WITH_PLAYBACK_CMD_RECORD: "u16",
    JD_SOUND_RECORDER_WITH_PLAYBACK_REG_STATUS: "u8",
    JD_SOUND_RECORDER_WITH_PLAYBACK_REG_TIME: "u16",
    JD_SOUND_RECORDER_WITH_PLAYBACK_REG_VOLUME: "u0.8"
}



import asyncio
from typing import Optional, Union, cast
from azure.iot.device.aio import IoTHubDeviceClient # pyright: ignore
from azure.iot.device import Message # pyright: ignore

from jacdac.pack import jdpack

from ..bus import Bus, Server
from ..packet import JDPacket
from .constants import *

CONNECTION_STRING_FILE = "./.jacdac/iothub.txt"


class AzureIotHubHealthServer(Server):
    """A Azure IoT Hub Health server

    This server uses the 'azure-iot-device' PyPi package

    pip install azure-iot-device
    """

    def __init__(self, bus: Bus) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_AZURE_IOT_HUB_HEALTH)
        self.device_client: Optional[IoTHubDeviceClient] = None
        self._connection_status = AzureIotHubHealthConnectionStatus.DISCONNECTED
        asyncio.create_task(self.connect())

    async def send_message(self, msg: Union[str, Message]):
        self.debug("send message")
        self._check_connection()
        if self.connection_status == AzureIotHubHealthConnectionStatus.DISCONNECTED:
            await self.connect()

        if self.connection_status == AzureIotHubHealthConnectionStatus.CONNECTED:
            try:
                await self.device_client.send_message(msg)  # type: ignore
                self.send_event(JD_AZURE_IOT_HUB_HEALTH_EV_MESSAGE_SENT)
            except Exception as e:
                print(e)
                self.error("send message error")
                await self.disconnect()

    @property
    def connection_status(self) -> AzureIotHubHealthConnectionStatus:
        return self._connection_status

    @connection_status.setter
    def connection_status(self, value: AzureIotHubHealthConnectionStatus):
        if self._connection_status != value:
            self._connection_status = value
            self.send_event(JD_AZURE_IOT_HUB_HEALTH_EV_CONNECTION_STATUS_CHANGE, jdpack(
                JD_AZURE_IOT_HUB_HEALTH_PACK_FORMATS[JD_AZURE_IOT_HUB_HEALTH_EV_CONNECTION_STATUS_CHANGE], self._connection_status))

    def _check_connection(self):
        if self.connection_status == AzureIotHubHealthConnectionStatus.CONNECTED and not self.device_client.connected:  # type: ignore
            self.connection_status = AzureIotHubHealthConnectionStatus.DISCONNECTED

    async def connect(self):
        if self.device_client:
            await self.disconnect()

        self.debug("connect")
        self.connection_status = AzureIotHubHealthConnectionStatus.CONNECTING
        try:
            with open(CONNECTION_STRING_FILE, "r") as f:
                conn_str = f.read()
                self.device_client = IoTHubDeviceClient.create_from_connection_string(  # type: ignore
                    conn_str)
                await self.device_client.connect()
            self.connection_status = AzureIotHubHealthConnectionStatus.CONNECTED
        except Exception as e:
            self.error("connect failed {}", e)
            self.connection_status = AzureIotHubHealthConnectionStatus.DISCONNECTED

    def set_connection_string(self, conn_str: str):
        self.debug("set connection string")
        with open(CONNECTION_STRING_FILE, "w") as f:
            f.write(conn_str)

    async def disconnect(self):
        if not self.device_client:
            return

        self.debug("disconnect")
        self.connection_status = AzureIotHubHealthConnectionStatus.DISCONNECTING
        try:
            await self.device_client.shutdown()
        except Exception as e:
            self.debug("connect failed {}", e)
        finally:
            self.device_client = None
            self.connection_status = AzureIotHubHealthConnectionStatus.DISCONNECTED

    def handle_packet(self, pkt: JDPacket):
        cmd = pkt.service_command

        if cmd == JD_GET(JD_AZURE_IOT_HUB_HEALTH_REG_HUB_NAME):
            self.send_report(pkt.not_implemented())
        elif cmd == JD_GET(JD_AZURE_IOT_HUB_HEALTH_REG_HUB_DEVICE_ID):
            self.send_report(pkt.not_implemented())
        elif cmd == JD_GET(JD_AZURE_IOT_HUB_HEALTH_REG_CONNECTION_STATUS):
            self.send_report(JDPacket.packed(
                cmd, JD_AZURE_IOT_HUB_HEALTH_PACK_FORMATS[JD_AZURE_IOT_HUB_HEALTH_REG_CONNECTION_STATUS], self.connection_status))
        elif cmd == JD_AZURE_IOT_HUB_HEALTH_CMD_CONNECT:
            self.handle_connect(pkt)
        elif cmd == JD_AZURE_IOT_HUB_HEALTH_CMD_DISCONNECT:
            self.handle_disconnect(pkt)
        elif cmd == JD_AZURE_IOT_HUB_HEALTH_CMD_SET_CONNECTION_STRING:
            self.handle_set_connection_string(pkt)

        return super().handle_packet(pkt)

    def handle_connect(self, pkt: JDPacket):
        self.debug("connect requested")
        asyncio.create_task(self.connect())

    def handle_disconnect(self, pkt: JDPacket):
        self.debug("disconnect requested")
        asyncio.create_task(self.disconnect())

    def handle_set_connection_string(self, pkt: JDPacket):
        [conn_str] = pkt.unpack("s")
        self.set_connection_string(cast(str, conn_str))



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client, EventHandlerFn, UnsubscribeFn
from .constants import *



class CodalMessageBusClient(Client):
    """
    A service that uses the [CODAL message bus](https://lancaster-university.github.io/microbit-docs/ubit/messageBus/) to send and receive small messages.
     * 
     * You can find known values for `source` in [CODAL repository](https://github.com/lancaster-university/codal-core/blob/master/inc/core/CodalComponent.h)
     * In MakeCode, you can listen for custom `source`, `value` values using [control.onEvent](https://makecode.microbit.org/reference/control/on-event].
    Implements a client for the `CODAL Message Bus <https://microsoft.github.io/jacdac-docs/services/codalmessagebus>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_CODAL_MESSAGE_BUS, JD_CODAL_MESSAGE_BUS_PACK_FORMATS, role)


    def on_message(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Raised by the server is triggered by the server. The filtering logic of which event to send over Jacdac is up to the server implementation.
        """
        return self.on_event(JD_CODAL_MESSAGE_BUS_EV_MESSAGE, handler)


    def send(self, source: int, value: int) -> None:
        """
        Send a message on the CODAL bus. If `source` is `0`, it is treated as wildcard.
        """
        self.send_cmd_packed(JD_CODAL_MESSAGE_BUS_CMD_SEND, source, value)
    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class WeightScaleClient(SensorClient):
    """
    A weight measuring sensor.
    Implements a client for the `Weight Scale <https://microsoft.github.io/jacdac-docs/services/weightscale>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_weight_value: float = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_WEIGHT_SCALE, JD_WEIGHT_SCALE_PACK_FORMATS, role)
        self.missing_weight_value = missing_weight_value

    @property
    def weight(self) -> Optional[float]:
        """
        The reported weight., _: kg
        """
        self.refresh_reading()
        return self.register(JD_WEIGHT_SCALE_REG_WEIGHT).value(self.missing_weight_value)

    @property
    def weight_error(self) -> Optional[float]:
        """
        (Optional) The estimate error on the reported reading., _: kg
        """
        return self.register(JD_WEIGHT_SCALE_REG_WEIGHT_ERROR).value()

    @property
    def zero_offset(self) -> Optional[float]:
        """
        (Optional) Calibrated zero offset error on the scale, i.e. the measured weight when nothing is on the scale.
        You do not need to subtract that from the reading, it has already been done., _: kg
        """
        return self.register(JD_WEIGHT_SCALE_REG_ZERO_OFFSET).value()

    @zero_offset.setter
    def zero_offset(self, value: float) -> None:
        self.register(JD_WEIGHT_SCALE_REG_ZERO_OFFSET).set_values(value)


    @property
    def gain(self) -> Optional[float]:
        """
        (Optional) Calibrated gain on the weight scale error., 
        """
        return self.register(JD_WEIGHT_SCALE_REG_GAIN).value()

    @gain.setter
    def gain(self, value: float) -> None:
        self.register(JD_WEIGHT_SCALE_REG_GAIN).set_values(value)


    @property
    def max_weight(self) -> Optional[float]:
        """
        (Optional) Maximum supported weight on the scale., _: kg
        """
        return self.register(JD_WEIGHT_SCALE_REG_MAX_WEIGHT).value()

    @property
    def min_weight(self) -> Optional[float]:
        """
        (Optional) Minimum recommend weight on the scale., _: kg
        """
        return self.register(JD_WEIGHT_SCALE_REG_MIN_WEIGHT).value()

    @property
    def weight_resolution(self) -> Optional[float]:
        """
        (Optional) Smallest, yet distinguishable change in reading., _: kg
        """
        return self.register(JD_WEIGHT_SCALE_REG_WEIGHT_RESOLUTION).value()

    @property
    def variant(self) -> Optional[WeightScaleVariant]:
        """
        (Optional) The type of physical scale, 
        """
        return self.register(JD_WEIGHT_SCALE_REG_VARIANT).value()


    def calibrate_zero_offset(self, ) -> None:
        """
        Call this command when there is nothing on the scale. If supported, the module should save the calibration data.
        """
        self.send_cmd_packed(JD_WEIGHT_SCALE_CMD_CALIBRATE_ZERO_OFFSET, )

    def calibrate_gain(self, weight: float) -> None:
        """
        Call this command with the weight of the thing on the scale.
        """
        self.send_cmd_packed(JD_WEIGHT_SCALE_CMD_CALIBRATE_GAIN, weight)
    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *
from typing import Optional


class BrailleDisplayClient(Client):
    """
    A Braille pattern display module. This module display [unicode braille patterns](https://www.unicode.org/charts/PDF/U2800.pdf), country specific encoding have to be implemented by the clients.
    Implements a client for the `Braille display <https://microsoft.github.io/jacdac-docs/services/brailledisplay>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_BRAILLE_DISPLAY, JD_BRAILLE_DISPLAY_PACK_FORMATS, role)


    @property
    def enabled(self) -> Optional[bool]:
        """
        Determines if the braille display is active., 
        """
        return self.register(JD_BRAILLE_DISPLAY_REG_ENABLED).bool_value()

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self.register(JD_BRAILLE_DISPLAY_REG_ENABLED).set_values(value)


    @property
    def patterns(self) -> Optional[str]:
        """
        Braille patterns to show. Must be unicode characters between `0x2800` and `0x28ff`., 
        """
        return self.register(JD_BRAILLE_DISPLAY_REG_PATTERNS).value()

    @patterns.setter
    def patterns(self, value: str) -> None:
        self.enabled = True
        self.register(JD_BRAILLE_DISPLAY_REG_PATTERNS).set_values(value)


    @property
    def length(self) -> Optional[int]:
        """
        Gets the number of patterns that can be displayed., _: #
        """
        return self.register(JD_BRAILLE_DISPLAY_REG_LENGTH).value()

    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class IlluminanceClient(SensorClient):
    """
    Detects the amount of light falling onto a given surface area.
     * 
     * Note that this is different from _luminance_, the amount of light that passes through, emits from, or reflects off an object.
    Implements a client for the `Illuminance <https://microsoft.github.io/jacdac-docs/services/illuminance>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_illuminance_value: float = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_ILLUMINANCE, JD_ILLUMINANCE_PACK_FORMATS, role)
        self.missing_illuminance_value = missing_illuminance_value

    @property
    def illuminance(self) -> Optional[float]:
        """
        The amount of illuminance, as lumens per square metre., _: lux
        """
        self.refresh_reading()
        return self.register(JD_ILLUMINANCE_REG_ILLUMINANCE).value(self.missing_illuminance_value)

    @property
    def illuminance_error(self) -> Optional[float]:
        """
        (Optional) Error on the reported sensor value., _: lux
        """
        return self.register(JD_ILLUMINANCE_REG_ILLUMINANCE_ERROR).value()

    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient, EventHandlerFn, UnsubscribeFn
from .constants import *
from typing import Optional, Tuple


class GamepadClient(SensorClient):
    """
    A two axis directional gamepad with optional buttons.
    Implements a client for the `Gamepad <https://microsoft.github.io/jacdac-docs/services/gamepad>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_direction_value: Tuple[GamepadButtons, float, float] = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_GAMEPAD, JD_GAMEPAD_PACK_FORMATS, role)
        self.missing_direction_value = missing_direction_value

    @property
    def direction(self) -> Optional[Tuple[GamepadButtons, float, float]]:
        """
        If the gamepad is analog, the directional buttons should be "simulated", based on gamepad position
        (`Left` is `{ x = -1, y = 0 }`, `Up` is `{ x = 0, y = -1}`).
        If the gamepad is digital, then each direction will read as either `-1`, `0`, or `1` (in fixed representation).
        The primary button on the gamepad is `A`., x: /,y: /
        """
        self.refresh_reading()
        return self.register(JD_GAMEPAD_REG_DIRECTION).value(self.missing_direction_value)

    @property
    def variant(self) -> Optional[GamepadVariant]:
        """
        (Optional) The type of physical gamepad., 
        """
        return self.register(JD_GAMEPAD_REG_VARIANT).value()

    @property
    def buttons_available(self) -> Optional[GamepadButtons]:
        """
        Indicates a bitmask of the buttons that are mounted on the gamepad.
        If the `Left`/`Up`/`Right`/`Down` buttons are marked as available here, the gamepad is digital.
        Even when marked as not available, they will still be simulated based on the analog gamepad., 
        """
        return self.register(JD_GAMEPAD_REG_BUTTONS_AVAILABLE).value()

    def on_buttons_changed(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted whenever the state of buttons changes.
        """
        return self.on_event(JD_GAMEPAD_EV_BUTTONS_CHANGED, handler)

    



import threading
from typing import List, Optional
from time import sleep
from tokenize import Number
from jacdac.bus import Transport
from jacdac.util import buf2hex, hex2buf, now
from gpiod import Chip, Line, LineBulk, LINE_REQ_EV_RISING_EDGE, LINE_REQ_FLAG_ACTIVE_LOW, LINE_REQ_DIR_OUT # type: ignore
from spidev import SpiDev # type: ignore
from weakref import finalize

XFER_SIZE = 256
# https://git.kernel.org/pub/scm/libs/libgpiod/libgpiod.git/tree/bindings/python/gpiodmodule.c?h=v1.6.x&id=27cacfe377114f6acf67cd943d1ca01bb30e0f2b
RPI_CHIP = 'pinctrl-bcm2835'
RPI_PIN_TX_READY = 24
RPI_PIN_RX_READY = 25
RPI_PIN_RST = 22
CONSUMER = "jacdac"
MAX_SEND_QUEUE_LEN = 10

class SpiTransport(Transport):
    def __init__(self):
        self.chip: Chip = None
        self.rxtx: LineBulk = None
        self.spi: SpiDev = None
        self.sendQueue: List[bytes] = []

        try:
            self._open()
        finally:
            self._finalizer = finalize(self, self._cleanup, self.chip, self.spi)

    def _open(self) -> None:
        self.sendQueue: List[bytes] = []
        print("spi: select chip")
        self.chip = Chip(RPI_CHIP)
        # monitor rx,tx in bulk
        print("spi: request rx,tx")
        self.rxtx = self.chip.get_lines([RPI_PIN_RX_READY, RPI_PIN_TX_READY])
        self.rxtx.request(consumer = CONSUMER, type = LINE_REQ_EV_RISING_EDGE, flags = LINE_REQ_FLAG_ACTIVE_LOW)
        self._flip_reset()
        print("spi: open device")
        self.spi = SpiDev()
        self.spi.open(0, 0)
        self.spi.max_speed_hz = 15600000
        self.spi.bits_per_word = 8
        print("spi: start read loop")
        t = threading.Thread(target=self._read_loop)
        t.start()

    @classmethod
    def _cleanup(cls, chip: Chip, spi: SpiDev):
        if not chip is None:
            try:
                chip.close()
                print("spi: chip closed")
            except:
                print("error: chip failed to close")

        if not spi is None:
            try:
                spi.close()
                print("spi: device closed")
            except:
                print("error: device failed to close")
    
    def close(self):
        print("spi: close")
        chip = self.chip
        spi = self.spi
        if not chip is None or not spi is None:
            self.chip = None
            self.spi = None
            self.rxtx = None
            self.sendQueue = []
            self._cleanup(chip, spi)
            self._finalizer.detach()
        
    def _flip_reset(self) -> None:
        print("spi: reset bridge")
        self.sendQueue = []
        rst = self.chip.get_line(RPI_PIN_RST)
        try:
            rst.request(consumer = CONSUMER, type = LINE_REQ_DIR_OUT)
            rst.set_value(0)
            sleep(0.01)
            rst.set_value(1)
        finally:
            rst.release()

    def send(self, pkt: bytes) -> None:
        self.sendQueue.append(pkt)
        self._transfer()

    def _read_loop(self) -> None:
        try:
            while True:
                ev_lines = self.rxtx.event_wait(nsec = 1) # List[Line]
                if ev_lines:
                    self._transfer()
        except:
            self.close()
            raise

    def _read_ready_pins(self) -> List[int]:
        rxtxv = self.rxtx.get_values() 
        if rxtxv is None:
            return [0, 0]
        else:
            return rxtxv

    def _transfer(self) -> None:
        if len(self.sendQueue) > MAX_SEND_QUEUE_LEN:
            self._flip_reset()

        try:
            while self._transfer_frame():
                pass
        except:
            self.close()
            raise

    def _transfer_frame(self) -> bool:
        [rx, tx] = self._read_ready_pins()
        rxReady = rx != 0
        txReady = tx != 0
        sendtx = txReady and len(self.sendQueue) > 0

      #  print("spi: transfer rx:" + str(rx) + ", tx: " + str(tx) + ", queue: " + str(len(self.sendQueue)))

        if not sendtx and not rxReady:
            return False

        # allocate transfer buffers
        txqueue = bytearray(XFER_SIZE)

        # assemble packets into send buffer
        txq_ptr = 0
        while txReady and len(self.sendQueue) > 0:
            pkt = self.sendQueue[0]
            npkt = len(pkt)
            if txq_ptr + npkt > len(txqueue):
                break
            self.sendQueue.pop(0)
            txqueue[txq_ptr:txq_ptr+npkt] = pkt
            txq_ptr += (npkt + 3) & ~3

        if txq_ptr == 0 and not rxReady:
            return False

        rxqueue = bytearray(0)
        if txq_ptr > 0:
            txqueue = bytearray(txqueue[0::txq_ptr])
            print(str(now()) + " " + buf2hex(txqueue) + " send frame")
            rxqueue = bytearray(self.spi.xfer2(txqueue))
        elif rxReady:
            rxqueue = bytearray(self.spi.readbytes(XFER_SIZE))
        if rxReady:
            if rxqueue is None:
                print("recv failed")
                return False
            #print(str(now()) + " " + buf2hex(rxqueue) + " recv frame")
            
            framep = 0
            while framep + 4 < len(rxqueue) :
                frame2 = rxqueue[framep + 2]
                if frame2 == 0:
                    # print("spi: empty frame")
                    break
                sz = frame2 + 12
                if framep + sz > len(rxqueue):
                    print("spi: frame size out of range")
                    break
                frame0 = rxqueue[framep]
                frame1 = rxqueue[framep + 1]
                frame3 = rxqueue[framep + 3]
                if frame0 == 0xff and frame1 == 0xff and frame3 == 0xff :
                    # skip bogus packet
                    print("spi: skip bogus pkt")
                    pass
                else:
                    buf = bytearray(rxqueue[framep:framep+sz])
                    print("spi: recv pkt " + buf2hex(buf))
                    if buf and self.on_receive:
                        self.on_receive(buf)
                sz = (sz + 3) & ~3
                framep += sz
        # and we're done
        return True



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *
from typing import Optional


class DmxClient(Client):
    """
    A service that can send DMX512-A packets with limited size. This service is designed to allow tinkering with a few DMX devices, but only allows 235 channels. More about DMX at https://en.wikipedia.org/wiki/DMX512.
    Implements a client for the `DMX <https://microsoft.github.io/jacdac-docs/services/dmx>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_DMX, JD_DMX_PACK_FORMATS, role)


    @property
    def enabled(self) -> Optional[bool]:
        """
        Determines if the DMX bridge is active., 
        """
        return self.register(JD_DMX_REG_ENABLED).bool_value()

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self.register(JD_DMX_REG_ENABLED).set_values(value)



    def send(self, channels: bytes) -> None:
        """
        Send a DMX packet, up to 236bytes long, including the start code.
        """
        self.send_cmd_packed(JD_DMX_CMD_SEND, channels)
    



# Autogenerated constants for Solenoid service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_SOLENOID = const(0x171723ca)


class SolenoidVariant(IntEnum):
    PUSH_PULL = const(0x1)
    VALVE = const(0x2)
    LATCH = const(0x3)


JD_SOLENOID_REG_PULLED = const(JD_REG_INTENSITY)
JD_SOLENOID_REG_VARIANT = const(JD_REG_VARIANT)
JD_SOLENOID_PACK_FORMATS = {
    JD_SOLENOID_REG_PULLED: "u8",
    JD_SOLENOID_REG_VARIANT: "u8"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *
from typing import Optional


class LightBulbClient(Client):
    """
    A light bulb controller.
    Implements a client for the `Light bulb <https://microsoft.github.io/jacdac-docs/services/lightbulb>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_LIGHT_BULB, JD_LIGHT_BULB_PACK_FORMATS, role)


    @property
    def brightness(self) -> Optional[float]:
        """
        Indicates the brightness of the light bulb. Zero means completely off and 0xffff means completely on.
        For non-dimmable lights, the value should be clamp to 0xffff for any non-zero value., _: /
        """
        return self.register(JD_LIGHT_BULB_REG_BRIGHTNESS).float_value(100)

    @brightness.setter
    def brightness(self, value: float) -> None:
        self.register(JD_LIGHT_BULB_REG_BRIGHTNESS).set_values(value / 100)


    @property
    def dimmable(self) -> Optional[bool]:
        """
        (Optional) Indicates if the light supports dimming., 
        """
        return self.register(JD_LIGHT_BULB_REG_DIMMABLE).bool_value()

    



# Autogenerated constants for TCP service
from enum import IntEnum
from jacdac.constants import *
JD_SERVICE_CLASS_TCP = const(0x1b43b70b)


class TcpTcpError(IntEnum):
    INVALID_COMMAND = const(0x1)
    INVALID_COMMAND_PAYLOAD = const(0x2)


JD_TCP_CMD_OPEN = const(0x80)
JD_TCP_PIPE_OPEN_SSL = const(0x1)
JD_TCP_PIPE_ERROR = const(0x0)
JD_TCP_PACK_FORMATS = {
    JD_TCP_CMD_OPEN: "b[12]",
    JD_TCP_PIPE_OPEN_SSL: "u16 s",
    JD_TCP_PIPE_ERROR: "i32"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class HumidityClient(SensorClient):
    """
    A sensor measuring humidity of outside environment.
    Implements a client for the `Humidity <https://microsoft.github.io/jacdac-docs/services/humidity>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_humidity_value: float = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_HUMIDITY, JD_HUMIDITY_PACK_FORMATS, role, preferred_interval = 5000)
        self.missing_humidity_value = missing_humidity_value

    @property
    def humidity(self) -> Optional[float]:
        """
        The relative humidity in percentage of full water saturation., _: %RH
        """
        self.refresh_reading()
        return self.register(JD_HUMIDITY_REG_HUMIDITY).value(self.missing_humidity_value)

    @property
    def humidity_error(self) -> Optional[float]:
        """
        (Optional) The real humidity is between `humidity - humidity_error` and `humidity + humidity_error`., _: %RH
        """
        return self.register(JD_HUMIDITY_REG_HUMIDITY_ERROR).value()

    @property
    def min_humidity(self) -> Optional[float]:
        """
        Lowest humidity that can be reported., _: %RH
        """
        return self.register(JD_HUMIDITY_REG_MIN_HUMIDITY).value()

    @property
    def max_humidity(self) -> Optional[float]:
        """
        Highest humidity that can be reported., _: %RH
        """
        return self.register(JD_HUMIDITY_REG_MAX_HUMIDITY).value()

    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client, EventHandlerFn, UnsubscribeFn
from .constants import *



class ArcadeGamepadClient(Client):
    """
    This service is deprecated in favor of `gamepad` (although it is currently used by the micro:bit Arcade smart shield).
     * A gamepad with direction and action buttons for one player.
     * If a device has multiple controllers, it should have multiple gamepad services, using consecutive service identifiers.
    Implements a client for the `Arcade Gamepad <https://microsoft.github.io/jacdac-docs/services/arcadegamepad>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_ARCADE_GAMEPAD, JD_ARCADE_GAMEPAD_PACK_FORMATS, role)


    def on_down(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when button goes from inactive to active.
        """
        return self.on_event(JD_ARCADE_GAMEPAD_EV_DOWN, handler)

    def on_up(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when button goes from active to inactive.
        """
        return self.on_event(JD_ARCADE_GAMEPAD_EV_UP, handler)

    



from ..bus import Bus
from .server import SoundPlayServer, create_playsound

if __name__ == '__main__':
    def main():
        bus = Bus()
        server = SoundPlayServer(
            bus, "./jacdac/sound_player/sounds", sound_player=create_playsound())
    main()



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client, EventHandlerFn, UnsubscribeFn
from .constants import *
from typing import Optional


class AzureIotHubHealthClient(Client):
    """
    Health and diagnostics information about the Azure Iot Hub connection.
    Implements a client for the `Azure IoT Hub Health <https://microsoft.github.io/jacdac-docs/services/azureiothubhealth>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_AZURE_IOT_HUB_HEALTH, JD_AZURE_IOT_HUB_HEALTH_PACK_FORMATS, role)


    @property
    def hub_name(self) -> Optional[str]:
        """
        (Optional) Something like `my-iot-hub.azure-devices.net` if available., 
        """
        return self.register(JD_AZURE_IOT_HUB_HEALTH_REG_HUB_NAME).value()

    @property
    def hub_device_id(self) -> Optional[str]:
        """
        (Optional) Device identifier in Azure Iot Hub if available., 
        """
        return self.register(JD_AZURE_IOT_HUB_HEALTH_REG_HUB_DEVICE_ID).value()

    @property
    def connection_status(self) -> Optional[AzureIotHubHealthConnectionStatus]:
        """
        Indicates the status of connection. A message beyond the [0..3] range represents an HTTP error code., 
        """
        return self.register(JD_AZURE_IOT_HUB_HEALTH_REG_CONNECTION_STATUS).value()

    @property
    def push_period(self) -> Optional[int]:
        """
        How often to push data to the cloud., _: ms
        """
        return self.register(JD_AZURE_IOT_HUB_HEALTH_REG_PUSH_PERIOD).value()

    @push_period.setter
    def push_period(self, value: int) -> None:
        self.register(JD_AZURE_IOT_HUB_HEALTH_REG_PUSH_PERIOD).set_values(value)


    @property
    def push_watchdog_period(self) -> Optional[int]:
        """
        If no message is published within given period, the device resets.
        This can be due to connectivity problems or due to the device having nothing to publish.
        Forced to be at least `2 * flush_period`.
        Set to `0` to disable (default)., _: ms
        """
        return self.register(JD_AZURE_IOT_HUB_HEALTH_REG_PUSH_WATCHDOG_PERIOD).value()

    @push_watchdog_period.setter
    def push_watchdog_period(self, value: int) -> None:
        self.register(JD_AZURE_IOT_HUB_HEALTH_REG_PUSH_WATCHDOG_PERIOD).set_values(value)


    def on_connection_status_change(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Raised when the connection status changes
        """
        return self.on_event(JD_AZURE_IOT_HUB_HEALTH_EV_CONNECTION_STATUS_CHANGE, handler)

    def on_message_sent(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Raised when a message has been sent to the hub.
        """
        return self.on_event(JD_AZURE_IOT_HUB_HEALTH_EV_MESSAGE_SENT, handler)

    



# Autogenerated constants for CODAL Message Bus service
from jacdac.constants import *
JD_SERVICE_CLASS_CODAL_MESSAGE_BUS = const(0x121ff81d)
JD_CODAL_MESSAGE_BUS_CMD_SEND = const(0x80)
JD_CODAL_MESSAGE_BUS_EV_MESSAGE = const(0x80)
JD_CODAL_MESSAGE_BUS_PACK_FORMATS = {
    JD_CODAL_MESSAGE_BUS_CMD_SEND: "u16 u16",
    JD_CODAL_MESSAGE_BUS_EV_MESSAGE: "u16 u16"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional, Tuple


class MagnetometerClient(SensorClient):
    """
    A 3-axis magnetometer.
    Implements a client for the `Magnetometer <https://microsoft.github.io/jacdac-docs/services/magnetomer>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_forces_value: Tuple[int, int, int] = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_MAGNETOMETER, JD_MAGNETOMETER_PACK_FORMATS, role)
        self.missing_forces_value = missing_forces_value

    @property
    def forces(self) -> Optional[Tuple[int, int, int]]:
        """
        Indicates the current magnetic field on magnetometer.
        For reference: `1 mgauss` is `100 nT` (and `1 gauss` is `100 000 nT`)., x: nT,y: nT,z: nT
        """
        self.refresh_reading()
        return self.register(JD_MAGNETOMETER_REG_FORCES).value(self.missing_forces_value)

    @property
    def forces_error(self) -> Optional[int]:
        """
        (Optional) Absolute estimated error on the readings., _: nT
        """
        return self.register(JD_MAGNETOMETER_REG_FORCES_ERROR).value()


    def calibrate(self, ) -> None:
        """
        Forces a calibration sequence where the user/device
        might have to rotate to be calibrated.
        """
        self.send_cmd_packed(JD_MAGNETOMETER_CMD_CALIBRATE, )
    



# Autogenerated constants for Button service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_BUTTON = const(0x1473a263)
JD_BUTTON_REG_PRESSURE = const(JD_REG_READING)
JD_BUTTON_REG_ANALOG = const(0x180)
JD_BUTTON_REG_PRESSED = const(0x181)
JD_BUTTON_EV_DOWN = const(JD_EV_ACTIVE)
JD_BUTTON_EV_UP = const(JD_EV_INACTIVE)
JD_BUTTON_EV_HOLD = const(0x81)
JD_BUTTON_PACK_FORMATS = {
    JD_BUTTON_REG_PRESSURE: "u0.16",
    JD_BUTTON_REG_ANALOG: "u8",
    JD_BUTTON_REG_PRESSED: "u8",
    JD_BUTTON_EV_UP: "u32",
    JD_BUTTON_EV_HOLD: "u32"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client, EventHandlerFn, UnsubscribeFn
from .constants import *
from typing import Optional, Tuple


class ProtoTestClient(Client):
    """
    This is test service to validate the protocol packet transmissions between the browser and a MCU.
     * Use this page if you are porting Jacdac to a new platform.
    Implements a client for the `Protocol Test <https://microsoft.github.io/jacdac-docs/services/prototest>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_PROTO_TEST, JD_PROTO_TEST_PACK_FORMATS, role)


    @property
    def rw_bool(self) -> Optional[bool]:
        """
        A read write bool register., 
        """
        return self.register(JD_PROTO_TEST_REG_RW_BOOL).bool_value()

    @rw_bool.setter
    def rw_bool(self, value: bool) -> None:
        self.register(JD_PROTO_TEST_REG_RW_BOOL).set_values(value)


    @property
    def ro_bool(self) -> Optional[bool]:
        """
        A read only bool register. Mirrors rw_bool., 
        """
        return self.register(JD_PROTO_TEST_REG_RO_BOOL).bool_value()

    @property
    def rw_u32(self) -> Optional[int]:
        """
        A read write u32 register., 
        """
        return self.register(JD_PROTO_TEST_REG_RW_U32).value()

    @rw_u32.setter
    def rw_u32(self, value: int) -> None:
        self.register(JD_PROTO_TEST_REG_RW_U32).set_values(value)


    @property
    def ro_u32(self) -> Optional[int]:
        """
        A read only u32 register.. Mirrors rw_u32., 
        """
        return self.register(JD_PROTO_TEST_REG_RO_U32).value()

    @property
    def rw_i32(self) -> Optional[int]:
        """
        A read write i32 register., 
        """
        return self.register(JD_PROTO_TEST_REG_RW_I32).value()

    @rw_i32.setter
    def rw_i32(self, value: int) -> None:
        self.register(JD_PROTO_TEST_REG_RW_I32).set_values(value)


    @property
    def ro_i32(self) -> Optional[int]:
        """
        A read only i32 register.. Mirrors rw_i32., 
        """
        return self.register(JD_PROTO_TEST_REG_RO_I32).value()

    @property
    def rw_string(self) -> Optional[str]:
        """
        A read write string register., 
        """
        return self.register(JD_PROTO_TEST_REG_RW_STRING).value()

    @rw_string.setter
    def rw_string(self, value: str) -> None:
        self.register(JD_PROTO_TEST_REG_RW_STRING).set_values(value)


    @property
    def ro_string(self) -> Optional[str]:
        """
        A read only string register. Mirrors rw_string., 
        """
        return self.register(JD_PROTO_TEST_REG_RO_STRING).value()

    @property
    def rw_bytes(self) -> Optional[bytes]:
        """
        A read write string register., 
        """
        return self.register(JD_PROTO_TEST_REG_RW_BYTES).value()

    @rw_bytes.setter
    def rw_bytes(self, value: bytes) -> None:
        self.register(JD_PROTO_TEST_REG_RW_BYTES).set_values(value)


    @property
    def ro_bytes(self) -> Optional[bytes]:
        """
        A read only string register. Mirrors ro_bytes., 
        """
        return self.register(JD_PROTO_TEST_REG_RO_BYTES).value()

    @property
    def rw_i8_u8_u16_i32(self) -> Optional[Tuple[int, int, int, int]]:
        """
        A read write i8, u8, u16, i32 register., 
        """
        return self.register(JD_PROTO_TEST_REG_RW_I8_U8_U16_I32).value()

    @rw_i8_u8_u16_i32.setter
    def rw_i8_u8_u16_i32(self, value: Tuple[int, int, int, int]) -> None:
        self.register(JD_PROTO_TEST_REG_RW_I8_U8_U16_I32).set_values(*value)


    @property
    def ro_i8_u8_u16_i32(self) -> Optional[Tuple[int, int, int, int]]:
        """
        A read only i8, u8, u16, i32 register.. Mirrors rw_i8_u8_u16_i32., 
        """
        return self.register(JD_PROTO_TEST_REG_RO_I8_U8_U16_I32).value()

    @property
    def rw_u8_string(self) -> Optional[Tuple[int, str]]:
        """
        A read write u8, string register., 
        """
        return self.register(JD_PROTO_TEST_REG_RW_U8_STRING).value()

    @rw_u8_string.setter
    def rw_u8_string(self, value: Tuple[int, str]) -> None:
        self.register(JD_PROTO_TEST_REG_RW_U8_STRING).set_values(*value)


    @property
    def ro_u8_string(self) -> Optional[Tuple[int, str]]:
        """
        A read only u8, string register.. Mirrors rw_u8_string., 
        """
        return self.register(JD_PROTO_TEST_REG_RO_U8_STRING).value()

    def on_e_bool(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        An event raised when rw_bool is modified
        """
        return self.on_event(JD_PROTO_TEST_EV_E_BOOL, handler)

    def on_e_u32(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        An event raised when rw_u32 is modified
        """
        return self.on_event(JD_PROTO_TEST_EV_E_U32, handler)

    def on_e_i32(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        An event raised when rw_i32 is modified
        """
        return self.on_event(JD_PROTO_TEST_EV_E_I32, handler)

    def on_e_string(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        An event raised when rw_string is modified
        """
        return self.on_event(JD_PROTO_TEST_EV_E_STRING, handler)

    def on_e_bytes(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        An event raised when rw_bytes is modified
        """
        return self.on_event(JD_PROTO_TEST_EV_E_BYTES, handler)

    def on_e_i8_u8_u16_i32(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        An event raised when rw_i8_u8_u16_i32 is modified
        """
        return self.on_event(JD_PROTO_TEST_EV_E_I8_U8_U16_I32, handler)

    def on_e_u8_string(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        An event raised when rw_u8_string is modified
        """
        return self.on_event(JD_PROTO_TEST_EV_E_U8_STRING, handler)


    def c_bool(self, bo: bool) -> None:
        """
        A command to set rw_bool.
        """
        self.send_cmd_packed(JD_PROTO_TEST_CMD_C_BOOL, bo)

    def c_u32(self, u32: int) -> None:
        """
        A command to set rw_u32.
        """
        self.send_cmd_packed(JD_PROTO_TEST_CMD_C_U32, u32)

    def c_i32(self, i32: int) -> None:
        """
        A command to set rw_i32.
        """
        self.send_cmd_packed(JD_PROTO_TEST_CMD_C_I32, i32)

    def c_string(self, str: str) -> None:
        """
        A command to set rw_string.
        """
        self.send_cmd_packed(JD_PROTO_TEST_CMD_C_STRING, str)

    def c_bytes(self, bytes: bytes) -> None:
        """
        A command to set rw_string.
        """
        self.send_cmd_packed(JD_PROTO_TEST_CMD_C_BYTES, bytes)

    def c_i8_u8_u16_i32(self, i8: int, u8: int, u16: int, i32: int) -> None:
        """
        A command to set rw_bytes.
        """
        self.send_cmd_packed(JD_PROTO_TEST_CMD_C_I8_U8_U16_I32, i8, u8, u16, i32)

    def c_u8_string(self, u8: int, str: str) -> None:
        """
        A command to set rw_u8_string.
        """
        self.send_cmd_packed(JD_PROTO_TEST_CMD_C_U8_STRING, u8, str)
    



# Autogenerated constants for Pulse Oximeter service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_PULSE_OXIMETER = const(0x10bb4eb6)
JD_PULSE_OXIMETER_REG_OXYGEN = const(JD_REG_READING)
JD_PULSE_OXIMETER_REG_OXYGEN_ERROR = const(JD_REG_READING_ERROR)
JD_PULSE_OXIMETER_PACK_FORMATS = {
    JD_PULSE_OXIMETER_REG_OXYGEN: "u8.8",
    JD_PULSE_OXIMETER_REG_OXYGEN_ERROR: "u8.8"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class FlexClient(SensorClient):
    """
    A bending or deflection sensor.
    Implements a client for the `Flex <https://microsoft.github.io/jacdac-docs/services/flex>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_bending_value: float = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_FLEX, JD_FLEX_PACK_FORMATS, role)
        self.missing_bending_value = missing_bending_value

    @property
    def bending(self) -> Optional[float]:
        """
        A measure of the bending., _: /
        """
        self.refresh_reading()
        return self.register(JD_FLEX_REG_BENDING).float_value(self.missing_bending_value, 100)

    @property
    def length(self) -> Optional[int]:
        """
        (Optional) Length of the flex sensor, _: mm
        """
        return self.register(JD_FLEX_REG_LENGTH).value()

    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *
from typing import Optional


class IndexedScreenClient(Client):
    """
    A screen with indexed colors.
     * 
     * This is often run over an SPI connection, not regular single-wire Jacdac.
    Implements a client for the `Indexed screen <https://microsoft.github.io/jacdac-docs/services/indexedscreen>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_INDEXED_SCREEN, JD_INDEXED_SCREEN_PACK_FORMATS, role)


    @property
    def brightness(self) -> Optional[float]:
        """
        Set backlight brightness.
        If set to `0` the display may go to sleep., _: /
        """
        return self.register(JD_INDEXED_SCREEN_REG_BRIGHTNESS).float_value(100)

    @brightness.setter
    def brightness(self, value: float) -> None:
        self.register(JD_INDEXED_SCREEN_REG_BRIGHTNESS).set_values(value / 100)


    @property
    def bits_per_pixel(self) -> Optional[int]:
        """
        Determines the number of palette entries.
        Typical values are 1, 2, 4, or 8., _: bit
        """
        return self.register(JD_INDEXED_SCREEN_REG_BITS_PER_PIXEL).value()

    @property
    def width(self) -> Optional[int]:
        """
        Screen width in "natural" orientation., _: px
        """
        return self.register(JD_INDEXED_SCREEN_REG_WIDTH).value()

    @property
    def height(self) -> Optional[int]:
        """
        Screen height in "natural" orientation., _: px
        """
        return self.register(JD_INDEXED_SCREEN_REG_HEIGHT).value()

    @property
    def width_major(self) -> Optional[bool]:
        """
        If true, consecutive pixels in the "width" direction are sent next to each other (this is typical for graphics cards).
        If false, consecutive pixels in the "height" direction are sent next to each other.
        For embedded screen controllers, this is typically true iff `width < height`
        (in other words, it's only true for portrait orientation screens).
        Some controllers may allow the user to change this (though the refresh order may not be optimal then).
        This is independent of the `rotation` register., 
        """
        return self.register(JD_INDEXED_SCREEN_REG_WIDTH_MAJOR).bool_value()

    @width_major.setter
    def width_major(self, value: bool) -> None:
        self.register(JD_INDEXED_SCREEN_REG_WIDTH_MAJOR).set_values(value)


    @property
    def up_sampling(self) -> Optional[int]:
        """
        Every pixel sent over wire is represented by `up_sampling x up_sampling` square of physical pixels.
        Some displays may allow changing this (which will also result in changes to `width` and `height`).
        Typical values are 1 and 2., _: px
        """
        return self.register(JD_INDEXED_SCREEN_REG_UP_SAMPLING).value()

    @up_sampling.setter
    def up_sampling(self, value: int) -> None:
        self.register(JD_INDEXED_SCREEN_REG_UP_SAMPLING).set_values(value)


    @property
    def rotation(self) -> Optional[int]:
        """
        Possible values are 0, 90, 180 and 270 only.
        Write to this register do not affect `width` and `height` registers,
        and may be ignored by some screens., _: °
        """
        return self.register(JD_INDEXED_SCREEN_REG_ROTATION).value()

    @rotation.setter
    def rotation(self, value: int) -> None:
        self.register(JD_INDEXED_SCREEN_REG_ROTATION).set_values(value)



    def start_update(self, x: int, y: int, width: int, height: int) -> None:
        """
        Sets the update window for subsequent `set_pixels` commands.
        """
        self.send_cmd_packed(JD_INDEXED_SCREEN_CMD_START_UPDATE, x, y, width, height)

    def set_pixels(self, pixels: bytes) -> None:
        """
        Set pixels in current window, according to current palette.
        Each "line" of data is aligned to a byte.
        """
        self.send_cmd_packed(JD_INDEXED_SCREEN_CMD_SET_PIXELS, pixels)
    



from typing import Callable, cast

from jacdac.pack import jdpack
from ..bus import Bus, OutPipe, Server
from .constants import *
from ..packet import JDPacket
from os import path, listdir


def create_playsound() -> Callable[[str], None]:
    from playsound import playsound

    def play(f: str):
        playsound(f, False)
    return play


def create_pygame_mixer() -> Callable[[str], None]:
    from pygame import mixer # pyright: ignore
    mixer.init()

    def play(f: str):
        snd = mixer.Sound(f)
        snd.play()
    return play


class SoundPlayServer(Server):
    """A sound player server implementation.

    This server uses pygame.

    pip install pygame

    """

    def __init__(self, bus: Bus, sound_dir: str, *, instance_name: str = None, sound_player: Callable[[str], None] = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_SOUND_PLAYER, instance_name=instance_name)
        self.sound_dir = sound_dir
        self.sound_player = sound_player
        self.volume = 0.2

    def handle_packet(self, pkt: JDPacket):
        self.volume = self.handle_reg(
            pkt, JD_SOUND_PLAYER_REG_VOLUME, JD_SOUND_PLAYER_PACK_FORMATS[JD_SOUND_PLAYER_REG_VOLUME], self.volume)

        cmd = pkt.service_command
        if cmd == JD_SOUND_PLAYER_CMD_PLAY:
            self.handle_play(pkt)
        elif cmd == JD_SOUND_PLAYER_CMD_LIST_SOUNDS:
            self.handle_list_sounds(pkt)
        else:
            super().handle_packet(pkt)

    def handle_play(self, pkt: JDPacket):
        name = cast(str, pkt.unpack(
            JD_SOUND_PLAYER_PACK_FORMATS[JD_SOUND_PLAYER_CMD_PLAY])[0])
        for f in listdir(self.sound_dir):
            if path.basename(f) == name:
                fp = path.join(self.sound_dir, f)
                if self.sound_player:
                    self.sound_player(fp)
                break

    def handle_list_sounds(self, pkt: JDPacket):
        pipe = OutPipe(self.bus, pkt)
        files = listdir(self.sound_dir)
        for f in files:
            if f.endswith(".mp3") or f.endswith(".wav"):
                pipe.write(bytearray(jdpack("u32 s", 0, path.basename(f))))
        pipe.close()



# Autogenerated constants for Protocol Test service
from jacdac.constants import *
JD_SERVICE_CLASS_PROTO_TEST = const(0x16c7466a)
JD_PROTO_TEST_REG_RW_BOOL = const(0x81)
JD_PROTO_TEST_REG_RO_BOOL = const(0x181)
JD_PROTO_TEST_REG_RW_U32 = const(0x82)
JD_PROTO_TEST_REG_RO_U32 = const(0x182)
JD_PROTO_TEST_REG_RW_I32 = const(0x83)
JD_PROTO_TEST_REG_RO_I32 = const(0x183)
JD_PROTO_TEST_REG_RW_STRING = const(0x84)
JD_PROTO_TEST_REG_RO_STRING = const(0x184)
JD_PROTO_TEST_REG_RW_BYTES = const(0x85)
JD_PROTO_TEST_REG_RO_BYTES = const(0x185)
JD_PROTO_TEST_REG_RW_I8_U8_U16_I32 = const(0x86)
JD_PROTO_TEST_REG_RO_I8_U8_U16_I32 = const(0x186)
JD_PROTO_TEST_REG_RW_U8_STRING = const(0x87)
JD_PROTO_TEST_REG_RO_U8_STRING = const(0x187)
JD_PROTO_TEST_EV_E_BOOL = const(0x81)
JD_PROTO_TEST_EV_E_U32 = const(0x82)
JD_PROTO_TEST_EV_E_I32 = const(0x83)
JD_PROTO_TEST_EV_E_STRING = const(0x84)
JD_PROTO_TEST_EV_E_BYTES = const(0x85)
JD_PROTO_TEST_EV_E_I8_U8_U16_I32 = const(0x86)
JD_PROTO_TEST_EV_E_U8_STRING = const(0x87)
JD_PROTO_TEST_CMD_C_BOOL = const(0x81)
JD_PROTO_TEST_CMD_C_U32 = const(0x82)
JD_PROTO_TEST_CMD_C_I32 = const(0x83)
JD_PROTO_TEST_CMD_C_STRING = const(0x84)
JD_PROTO_TEST_CMD_C_BYTES = const(0x85)
JD_PROTO_TEST_CMD_C_I8_U8_U16_I32 = const(0x86)
JD_PROTO_TEST_CMD_C_U8_STRING = const(0x87)
JD_PROTO_TEST_CMD_C_REPORT_PIPE = const(0x90)
JD_PROTO_TEST_PACK_FORMATS = {
    JD_PROTO_TEST_REG_RW_BOOL: "u8",
    JD_PROTO_TEST_REG_RO_BOOL: "u8",
    JD_PROTO_TEST_REG_RW_U32: "u32",
    JD_PROTO_TEST_REG_RO_U32: "u32",
    JD_PROTO_TEST_REG_RW_I32: "i32",
    JD_PROTO_TEST_REG_RO_I32: "i32",
    JD_PROTO_TEST_REG_RW_STRING: "s",
    JD_PROTO_TEST_REG_RO_STRING: "s",
    JD_PROTO_TEST_REG_RW_BYTES: "b",
    JD_PROTO_TEST_REG_RO_BYTES: "b",
    JD_PROTO_TEST_REG_RW_I8_U8_U16_I32: "i8 u8 u16 i32",
    JD_PROTO_TEST_REG_RO_I8_U8_U16_I32: "i8 u8 u16 i32",
    JD_PROTO_TEST_REG_RW_U8_STRING: "u8 s",
    JD_PROTO_TEST_REG_RO_U8_STRING: "u8 s",
    JD_PROTO_TEST_EV_E_BOOL: "u8",
    JD_PROTO_TEST_EV_E_U32: "u32",
    JD_PROTO_TEST_EV_E_I32: "i32",
    JD_PROTO_TEST_EV_E_STRING: "s",
    JD_PROTO_TEST_EV_E_BYTES: "b",
    JD_PROTO_TEST_EV_E_I8_U8_U16_I32: "i8 u8 u16 i32",
    JD_PROTO_TEST_EV_E_U8_STRING: "u8 s",
    JD_PROTO_TEST_CMD_C_BOOL: "u8",
    JD_PROTO_TEST_CMD_C_U32: "u32",
    JD_PROTO_TEST_CMD_C_I32: "i32",
    JD_PROTO_TEST_CMD_C_STRING: "s",
    JD_PROTO_TEST_CMD_C_BYTES: "b",
    JD_PROTO_TEST_CMD_C_I8_U8_U16_I32: "i8 u8 u16 i32",
    JD_PROTO_TEST_CMD_C_U8_STRING: "u8 s",
    JD_PROTO_TEST_CMD_C_REPORT_PIPE: "b[12]"
}



# Autogenerated constants for Color service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_COLOR = const(0x1630d567)
JD_COLOR_REG_COLOR = const(JD_REG_READING)
JD_COLOR_PACK_FORMATS = {
    JD_COLOR_REG_COLOR: "u0.16 u0.16 u0.16"
}



# Autogenerated constants for Multitouch service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_MULTITOUCH = const(0x1d112ab5)
JD_MULTITOUCH_REG_CAPACITY = const(JD_REG_READING)
JD_MULTITOUCH_EV_TOUCH = const(JD_EV_ACTIVE)
JD_MULTITOUCH_EV_RELEASE = const(JD_EV_INACTIVE)
JD_MULTITOUCH_EV_TAP = const(0x80)
JD_MULTITOUCH_EV_LONG_PRESS = const(0x81)
JD_MULTITOUCH_EV_SWIPE_POS = const(0x90)
JD_MULTITOUCH_EV_SWIPE_NEG = const(0x91)
JD_MULTITOUCH_PACK_FORMATS = {
    JD_MULTITOUCH_REG_CAPACITY: "r: i16",
    JD_MULTITOUCH_EV_TOUCH: "u8",
    JD_MULTITOUCH_EV_RELEASE: "u8",
    JD_MULTITOUCH_EV_TAP: "u8",
    JD_MULTITOUCH_EV_LONG_PRESS: "u8",
    JD_MULTITOUCH_EV_SWIPE_POS: "u16 u8 u8",
    JD_MULTITOUCH_EV_SWIPE_NEG: "u16 u8 u8"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient, EventHandlerFn, UnsubscribeFn
from .constants import *
from typing import Optional, Tuple


class AccelerometerClient(SensorClient):
    """
    A 3-axis accelerometer.
    Implements a client for the `Accelerometer <https://microsoft.github.io/jacdac-docs/services/accelerometer>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_forces_value: Tuple[float, float, float] = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_ACCELEROMETER, JD_ACCELEROMETER_PACK_FORMATS, role)
        self.missing_forces_value = missing_forces_value

    @property
    def forces(self) -> Optional[Tuple[float, float, float]]:
        """
        Indicates the current forces acting on accelerometer., x: g,y: g,z: g
        """
        self.refresh_reading()
        return self.register(JD_ACCELEROMETER_REG_FORCES).value(self.missing_forces_value)

    @property
    def forces_error(self) -> Optional[float]:
        """
        (Optional) Error on the reading value., _: g
        """
        return self.register(JD_ACCELEROMETER_REG_FORCES_ERROR).value()

    @property
    def max_force(self) -> Optional[float]:
        """
        (Optional) Configures the range forces detected.
        The value will be "rounded up" to one of `max_forces_supported`., _: g
        """
        return self.register(JD_ACCELEROMETER_REG_MAX_FORCE).value()

    @max_force.setter
    def max_force(self, value: float) -> None:
        self.register(JD_ACCELEROMETER_REG_MAX_FORCE).set_values(value)


    def on_tilt_up(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when accelerometer is tilted in the given direction.
        """
        return self.on_event(JD_ACCELEROMETER_EV_TILT_UP, handler)

    def on_tilt_down(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when accelerometer is tilted in the given direction.
        """
        return self.on_event(JD_ACCELEROMETER_EV_TILT_DOWN, handler)

    def on_tilt_left(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when accelerometer is tilted in the given direction.
        """
        return self.on_event(JD_ACCELEROMETER_EV_TILT_LEFT, handler)

    def on_tilt_right(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when accelerometer is tilted in the given direction.
        """
        return self.on_event(JD_ACCELEROMETER_EV_TILT_RIGHT, handler)

    def on_face_up(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when accelerometer is laying flat in the given direction.
        """
        return self.on_event(JD_ACCELEROMETER_EV_FACE_UP, handler)

    def on_face_down(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when accelerometer is laying flat in the given direction.
        """
        return self.on_event(JD_ACCELEROMETER_EV_FACE_DOWN, handler)

    def on_freefall(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when total force acting on accelerometer is much less than 1g.
        """
        return self.on_event(JD_ACCELEROMETER_EV_FREEFALL, handler)

    def on_shake(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when forces change violently a few times.
        """
        return self.on_event(JD_ACCELEROMETER_EV_SHAKE, handler)

    def on_force_2g(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when force in any direction exceeds given threshold.
        """
        return self.on_event(JD_ACCELEROMETER_EV_FORCE_2G, handler)

    def on_force_3g(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when force in any direction exceeds given threshold.
        """
        return self.on_event(JD_ACCELEROMETER_EV_FORCE_3G, handler)

    def on_force_6g(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when force in any direction exceeds given threshold.
        """
        return self.on_event(JD_ACCELEROMETER_EV_FORCE_6G, handler)

    def on_force_8g(self, handler: EventHandlerFn) -> UnsubscribeFn:
        """
        Emitted when force in any direction exceeds given threshold.
        """
        return self.on_event(JD_ACCELEROMETER_EV_FORCE_8G, handler)

    



# Autogenerated constants for Real time clock service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_REAL_TIME_CLOCK = const(0x1a8b1a28)


class RealTimeClockVariant(IntEnum):
    COMPUTER = const(0x1)
    CRYSTAL = const(0x2)
    CUCKOO = const(0x3)


JD_REAL_TIME_CLOCK_REG_LOCAL_TIME = const(JD_REG_READING)
JD_REAL_TIME_CLOCK_REG_DRIFT = const(0x180)
JD_REAL_TIME_CLOCK_REG_PRECISION = const(0x181)
JD_REAL_TIME_CLOCK_REG_VARIANT = const(JD_REG_VARIANT)
JD_REAL_TIME_CLOCK_CMD_SET_TIME = const(0x80)
JD_REAL_TIME_CLOCK_PACK_FORMATS = {
    JD_REAL_TIME_CLOCK_REG_LOCAL_TIME: "u16 u8 u8 u8 u8 u8 u8",
    JD_REAL_TIME_CLOCK_REG_DRIFT: "u16.16",
    JD_REAL_TIME_CLOCK_REG_PRECISION: "u16.16",
    JD_REAL_TIME_CLOCK_REG_VARIANT: "u8",
    JD_REAL_TIME_CLOCK_CMD_SET_TIME: "u16 u8 u8 u8 u8 u8 u8"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class SensorAggregatorClient(SensorClient):
    """
    Aggregate data from multiple sensors into a single stream
     * (often used as input to machine learning models on the same device, see model runner service).
    Implements a client for the `Sensor Aggregator <https://microsoft.github.io/jacdac-docs/services/sensoraggregator>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_current_sample_value: bytes = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_SENSOR_AGGREGATOR, JD_SENSOR_AGGREGATOR_PACK_FORMATS, role)
        self.missing_current_sample_value = missing_current_sample_value

    @property
    def num_samples(self) -> Optional[int]:
        """
        Number of input samples collected so far., 
        """
        return self.register(JD_SENSOR_AGGREGATOR_REG_NUM_SAMPLES).value()

    @property
    def sample_size(self) -> Optional[int]:
        """
        Size of a single sample., _: B
        """
        return self.register(JD_SENSOR_AGGREGATOR_REG_SAMPLE_SIZE).value()

    @property
    def streaming_samples(self) -> Optional[int]:
        """
        When set to `N`, will stream `N` samples as `current_sample` reading., _: #
        """
        return self.register(JD_SENSOR_AGGREGATOR_REG_STREAMING_SAMPLES).value()

    @streaming_samples.setter
    def streaming_samples(self, value: int) -> None:
        self.register(JD_SENSOR_AGGREGATOR_REG_STREAMING_SAMPLES).set_values(value)


    @property
    def current_sample(self) -> Optional[bytes]:
        """
        Last collected sample., 
        """
        self.refresh_reading()
        return self.register(JD_SENSOR_AGGREGATOR_REG_CURRENT_SAMPLE).value(self.missing_current_sample_value)

    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *
from typing import Optional


class SevenSegmentDisplayClient(Client):
    """
    A 7-segment numeric display, with one or more digits.
    Implements a client for the `7-segment display <https://microsoft.github.io/jacdac-docs/services/sevensegmentdisplay>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_SEVEN_SEGMENT_DISPLAY, JD_SEVEN_SEGMENT_DISPLAY_PACK_FORMATS, role)


    @property
    def digits(self) -> Optional[bytes]:
        """
        Each byte encodes the display status of a digit using,
        where lowest bit 0 encodes segment `A`, bit 1 encodes segments `B`, ..., bit 6 encodes segments `G`, and bit 7 encodes the decimal point (if present).
        If incoming `digits` data is smaller than `digit_count`, the remaining digits will be cleared.
        Thus, sending an empty `digits` payload clears the screen.
        
        ```text
        GFEDCBA DP
         - A -
         F   B
         |   |
         - G -
         |   |
         E   C   _
         |   |  |DP|
         - D -   -
        ```, 
        """
        return self.register(JD_SEVEN_SEGMENT_DISPLAY_REG_DIGITS).value()

    @digits.setter
    def digits(self, value: bytes) -> None:
        self.register(JD_SEVEN_SEGMENT_DISPLAY_REG_DIGITS).set_values(value)


    @property
    def brightness(self) -> Optional[float]:
        """
        (Optional) Controls the brightness of the LEDs. `0` means off., _: /
        """
        return self.register(JD_SEVEN_SEGMENT_DISPLAY_REG_BRIGHTNESS).float_value(100)

    @brightness.setter
    def brightness(self, value: float) -> None:
        self.register(JD_SEVEN_SEGMENT_DISPLAY_REG_BRIGHTNESS).set_values(value / 100)


    @property
    def double_dots(self) -> Optional[bool]:
        """
        (Optional) Turn on or off the column LEDs (separating minutes from hours, etc.) in of the segment.
        If the column LEDs is not supported, the value remains false., 
        """
        return self.register(JD_SEVEN_SEGMENT_DISPLAY_REG_DOUBLE_DOTS).bool_value()

    @double_dots.setter
    def double_dots(self, value: bool) -> None:
        self.register(JD_SEVEN_SEGMENT_DISPLAY_REG_DOUBLE_DOTS).set_values(value)


    @property
    def digit_count(self) -> Optional[int]:
        """
        The number of digits available on the display., 
        """
        return self.register(JD_SEVEN_SEGMENT_DISPLAY_REG_DIGIT_COUNT).value()

    @property
    def decimal_point(self) -> Optional[bool]:
        """
        (Optional) True if decimal points are available (on all digits)., 
        """
        return self.register(JD_SEVEN_SEGMENT_DISPLAY_REG_DECIMAL_POINT).bool_value()


    def set_number(self, value: float) -> None:
        """
        Shows the number on the screen using the decimal dot if available.
        """
        # TODO: implement client command
        raise RuntimeError("client command not implemented")
    



# Autogenerated constants for Jacscript Cloud service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_JACSCRIPT_CLOUD = const(0x14606e9c)


class JacscriptCloudCommandStatus(IntEnum):
    OK = const(0xc8)
    NOT_FOUND = const(0x194)
    BUSY = const(0x1ad)


JD_JACSCRIPT_CLOUD_CMD_UPLOAD = const(0x80)
JD_JACSCRIPT_CLOUD_CMD_UPLOAD_BIN = const(0x81)
JD_JACSCRIPT_CLOUD_CMD_ACK_CLOUD_COMMAND = const(0x83)
JD_JACSCRIPT_CLOUD_REG_CONNECTED = const(0x180)
JD_JACSCRIPT_CLOUD_REG_CONNECTION_NAME = const(0x181)
JD_JACSCRIPT_CLOUD_EV_CLOUD_COMMAND = const(0x81)
JD_JACSCRIPT_CLOUD_EV_CHANGE = const(JD_EV_CHANGE)
JD_JACSCRIPT_CLOUD_PACK_FORMATS = {
    JD_JACSCRIPT_CLOUD_CMD_UPLOAD: "z r: f64",
    JD_JACSCRIPT_CLOUD_CMD_UPLOAD_BIN: "b",
    JD_JACSCRIPT_CLOUD_CMD_ACK_CLOUD_COMMAND: "u32 u32 r: f64",
    JD_JACSCRIPT_CLOUD_REG_CONNECTED: "u8",
    JD_JACSCRIPT_CLOUD_REG_CONNECTION_NAME: "s",
    JD_JACSCRIPT_CLOUD_EV_CLOUD_COMMAND: "u32 z r: f64"
}



from ..bus import Bus
from .server import AzureIotHubHealthServer
from time import sleep
import asyncio
import tracemalloc

if __name__ == '__main__':
    async def main():
        tracemalloc.start()
        bus = Bus()
        hub_server = AzureIotHubHealthServer(bus)
        while True:
            sleep(1)
            await hub_server.send_message("hello")
    asyncio.run(main())



# Autogenerated constants for Role Manager service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_ROLE_MANAGER = const(0x1e4b7e66)
JD_ROLE_MANAGER_REG_AUTO_BIND = const(0x80)
JD_ROLE_MANAGER_REG_ALL_ROLES_ALLOCATED = const(0x181)
JD_ROLE_MANAGER_CMD_SET_ROLE = const(0x81)
JD_ROLE_MANAGER_CMD_CLEAR_ALL_ROLES = const(0x84)
JD_ROLE_MANAGER_CMD_LIST_ROLES = const(0x83)
JD_ROLE_MANAGER_EV_CHANGE = const(JD_EV_CHANGE)
JD_ROLE_MANAGER_PACK_FORMATS = {
    JD_ROLE_MANAGER_REG_AUTO_BIND: "u8",
    JD_ROLE_MANAGER_REG_ALL_ROLES_ALLOCATED: "u8",
    JD_ROLE_MANAGER_CMD_SET_ROLE: "b[8] u8 s",
    JD_ROLE_MANAGER_CMD_LIST_ROLES: "b[12]"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class UvIndexClient(SensorClient):
    """
    The UV Index is a measure of the intensity of ultraviolet (UV) rays from the Sun.
    Implements a client for the `UV index <https://microsoft.github.io/jacdac-docs/services/uvindex>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_uv_index_value: float = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_UV_INDEX, JD_UV_INDEX_PACK_FORMATS, role, preferred_interval = 1000)
        self.missing_uv_index_value = missing_uv_index_value

    @property
    def uv_index(self) -> Optional[float]:
        """
        Ultraviolet index, typically refreshed every second., _: uv
        """
        self.refresh_reading()
        return self.register(JD_UV_INDEX_REG_UV_INDEX).value(self.missing_uv_index_value)

    @property
    def uv_index_error(self) -> Optional[float]:
        """
        (Optional) Error on the UV measure., _: uv
        """
        return self.register(JD_UV_INDEX_REG_UV_INDEX_ERROR).value()

    @property
    def variant(self) -> Optional[UvIndexVariant]:
        """
        (Optional) The type of physical sensor and capabilities., 
        """
        return self.register(JD_UV_INDEX_REG_VARIANT).value()

    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class SoundLevelClient(SensorClient):
    """
    A sound level detector sensor, gives a relative indication of the sound level.
    Implements a client for the `Sound level <https://microsoft.github.io/jacdac-docs/services/soundlevel>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_sound_level_value: float = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_SOUND_LEVEL, JD_SOUND_LEVEL_PACK_FORMATS, role)
        self.missing_sound_level_value = missing_sound_level_value

    @property
    def sound_level(self) -> Optional[float]:
        """
        The sound level detected by the microphone, _: /
        """
        self.refresh_reading()
        return self.register(JD_SOUND_LEVEL_REG_SOUND_LEVEL).float_value(self.missing_sound_level_value, 100)

    @property
    def enabled(self) -> Optional[bool]:
        """
        Turn on or off the microphone., 
        """
        return self.register(JD_SOUND_LEVEL_REG_ENABLED).bool_value()

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self.register(JD_SOUND_LEVEL_REG_ENABLED).set_values(value)


    



# Autogenerated constants for WIFI service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_WIFI = const(0x18aae1fa)


class WifiAPFlags(IntEnum):
    HAS_PASSWORD = const(0x1)
    WPS = const(0x2)
    HAS_SECONDARY_CHANNEL_ABOVE = const(0x4)
    HAS_SECONDARY_CHANNEL_BELOW = const(0x8)
    IEEE_802_11B = const(0x100)
    IEEE_802_11A = const(0x200)
    IEEE_802_11G = const(0x400)
    IEEE_802_11N = const(0x800)
    IEEE_802_11AC = const(0x1000)
    IEEE_802_11AX = const(0x2000)
    IEEE_802_LONG_RANGE = const(0x8000)


JD_WIFI_CMD_LAST_SCAN_RESULTS = const(0x80)
JD_WIFI_CMD_ADD_NETWORK = const(0x81)
JD_WIFI_CMD_RECONNECT = const(0x82)
JD_WIFI_CMD_FORGET_NETWORK = const(0x83)
JD_WIFI_CMD_FORGET_ALL_NETWORKS = const(0x84)
JD_WIFI_CMD_SET_NETWORK_PRIORITY = const(0x85)
JD_WIFI_CMD_SCAN = const(0x86)
JD_WIFI_CMD_LIST_KNOWN_NETWORKS = const(0x87)
JD_WIFI_REG_RSSI = const(JD_REG_READING)
JD_WIFI_REG_ENABLED = const(JD_REG_INTENSITY)
JD_WIFI_REG_IP_ADDRESS = const(0x181)
JD_WIFI_REG_EUI_48 = const(0x182)
JD_WIFI_REG_SSID = const(0x183)
JD_WIFI_EV_GOT_IP = const(JD_EV_ACTIVE)
JD_WIFI_EV_LOST_IP = const(JD_EV_INACTIVE)
JD_WIFI_EV_SCAN_COMPLETE = const(0x80)
JD_WIFI_EV_NETWORKS_CHANGED = const(0x81)
JD_WIFI_EV_CONNECTION_FAILED = const(0x82)
JD_WIFI_PACK_FORMATS = {
    JD_WIFI_CMD_LAST_SCAN_RESULTS: "b[12]",
    JD_WIFI_CMD_ADD_NETWORK: "z z",
    JD_WIFI_CMD_FORGET_NETWORK: "s",
    JD_WIFI_CMD_SET_NETWORK_PRIORITY: "i16 s",
    JD_WIFI_CMD_LIST_KNOWN_NETWORKS: "b[12]",
    JD_WIFI_REG_RSSI: "i8",
    JD_WIFI_REG_ENABLED: "u8",
    JD_WIFI_REG_IP_ADDRESS: "b[16]",
    JD_WIFI_REG_EUI_48: "b[6]",
    JD_WIFI_REG_SSID: "s[32]",
    JD_WIFI_EV_SCAN_COMPLETE: "u16 u16",
    JD_WIFI_EV_CONNECTION_FAILED: "s"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *
from typing import Optional


class SolenoidClient(Client):
    """
    A push-pull solenoid is a type of relay that pulls a coil when activated.
    Implements a client for the `Solenoid <https://microsoft.github.io/jacdac-docs/services/solenoid>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_SOLENOID, JD_SOLENOID_PACK_FORMATS, role)


    @property
    def pulled(self) -> Optional[bool]:
        """
        Indicates whether the solenoid is energized and pulled (on) or pushed (off)., 
        """
        return self.register(JD_SOLENOID_REG_PULLED).bool_value()

    @pulled.setter
    def pulled(self, value: bool) -> None:
        self.register(JD_SOLENOID_REG_PULLED).set_values(value)


    @property
    def variant(self) -> Optional[SolenoidVariant]:
        """
        (Optional) Describes the type of solenoid used., 
        """
        return self.register(JD_SOLENOID_REG_VARIANT).value()

    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class CompassClient(SensorClient):
    """
    A sensor that measures the heading.
    Implements a client for the `Compass <https://microsoft.github.io/jacdac-docs/services/compass>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_heading_value: float = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_COMPASS, JD_COMPASS_PACK_FORMATS, role, preferred_interval = 1000)
        self.missing_heading_value = missing_heading_value

    @property
    def heading(self) -> Optional[float]:
        """
        The heading with respect to the magnetic north., _: °
        """
        self.refresh_reading()
        return self.register(JD_COMPASS_REG_HEADING).value(self.missing_heading_value)

    @property
    def enabled(self) -> Optional[bool]:
        """
        Turn on or off the sensor. Turning on the sensor may start a calibration sequence., 
        """
        return self.register(JD_COMPASS_REG_ENABLED).bool_value()

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self.register(JD_COMPASS_REG_ENABLED).set_values(value)


    @property
    def heading_error(self) -> Optional[float]:
        """
        (Optional) Error on the heading reading, _: °
        """
        return self.register(JD_COMPASS_REG_HEADING_ERROR).value()


    def calibrate(self, ) -> None:
        """
        Starts a calibration sequence for the compass.
        """
        self.send_cmd_packed(JD_COMPASS_CMD_CALIBRATE, )
    



# Autogenerated constants for Arcade Gamepad service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_ARCADE_GAMEPAD = const(0x1deaa06e)


class ArcadeGamepadButton(IntEnum):
    LEFT = const(0x1)
    UP = const(0x2)
    RIGHT = const(0x3)
    DOWN = const(0x4)
    A = const(0x5)
    B = const(0x6)
    MENU = const(0x7)
    SELECT = const(0x8)
    RESET = const(0x9)
    EXIT = const(0xa)


JD_ARCADE_GAMEPAD_REG_BUTTONS = const(JD_REG_READING)
JD_ARCADE_GAMEPAD_REG_AVAILABLE_BUTTONS = const(0x180)
JD_ARCADE_GAMEPAD_EV_DOWN = const(JD_EV_ACTIVE)
JD_ARCADE_GAMEPAD_EV_UP = const(JD_EV_INACTIVE)
JD_ARCADE_GAMEPAD_PACK_FORMATS = {
    JD_ARCADE_GAMEPAD_REG_BUTTONS: "r: u8 u0.8",
    JD_ARCADE_GAMEPAD_REG_AVAILABLE_BUTTONS: "r: u8",
    JD_ARCADE_GAMEPAD_EV_DOWN: "u8",
    JD_ARCADE_GAMEPAD_EV_UP: "u8"
}



# Autogenerated constants for Bootloader service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_BOOTLOADER = const(0x1ffa9948)


class BootloaderError(IntEnum):
    NO_ERROR = const(0x0)
    PACKET_TOO_SMALL = const(0x1)
    OUT_OF_FLASHABLE_RANGE = const(0x2)
    INVALID_PAGE_OFFSET = const(0x3)
    NOT_PAGE_ALIGNED = const(0x4)


JD_BOOTLOADER_CMD_INFO = const(JD_CMD_ANNOUNCE)
JD_BOOTLOADER_CMD_SET_SESSION = const(0x81)
JD_BOOTLOADER_CMD_PAGE_DATA = const(0x80)
JD_BOOTLOADER_PACK_FORMATS = {
    JD_BOOTLOADER_CMD_SET_SESSION: "u32",
    JD_BOOTLOADER_CMD_PAGE_DATA: "u32 u16 u8 u8 u32 u32 u32 u32 u32 b[208]"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional, Tuple


class GyroscopeClient(SensorClient):
    """
    A 3-axis gyroscope.
    Implements a client for the `Gyroscope <https://microsoft.github.io/jacdac-docs/services/gyroscope>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_rotation_rates_value: Tuple[float, float, float] = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_GYROSCOPE, JD_GYROSCOPE_PACK_FORMATS, role)
        self.missing_rotation_rates_value = missing_rotation_rates_value

    @property
    def rotation_rates(self) -> Optional[Tuple[float, float, float]]:
        """
        Indicates the current rates acting on gyroscope., x: °/s,y: °/s,z: °/s
        """
        self.refresh_reading()
        return self.register(JD_GYROSCOPE_REG_ROTATION_RATES).value(self.missing_rotation_rates_value)

    @property
    def rotation_rates_error(self) -> Optional[float]:
        """
        (Optional) Error on the reading value., _: °/s
        """
        return self.register(JD_GYROSCOPE_REG_ROTATION_RATES_ERROR).value()

    @property
    def max_rate(self) -> Optional[float]:
        """
        (Optional) Configures the range of rotation rates.
        The value will be "rounded up" to one of `max_rates_supported`., _: °/s
        """
        return self.register(JD_GYROSCOPE_REG_MAX_RATE).value()

    @max_rate.setter
    def max_rate(self, value: float) -> None:
        self.register(JD_GYROSCOPE_REG_MAX_RATE).set_values(value)


    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional


class ReflectedLightClient(SensorClient):
    """
    A sensor that detects light and dark surfaces, commonly used for line following robots.
    Implements a client for the `Reflected light <https://microsoft.github.io/jacdac-docs/services/reflectedlight>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_brightness_value: float = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_REFLECTED_LIGHT, JD_REFLECTED_LIGHT_PACK_FORMATS, role)
        self.missing_brightness_value = missing_brightness_value

    @property
    def brightness(self) -> Optional[float]:
        """
        Reports the reflected brightness. It may be a digital value or, for some sensor, analog value., _: /
        """
        self.refresh_reading()
        return self.register(JD_REFLECTED_LIGHT_REG_BRIGHTNESS).float_value(self.missing_brightness_value, 100)

    @property
    def variant(self) -> Optional[ReflectedLightVariant]:
        """
        (Optional) Type of physical sensor used, 
        """
        return self.register(JD_REFLECTED_LIGHT_REG_VARIANT).value()

    



# Autogenerated constants for Control service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_CONTROL = const(0x0)


class ControlAnnounceFlags(IntEnum):
    RESTART_COUNTER_STEADY = const(0xf)
    RESTART_COUNTER1 = const(0x1)
    RESTART_COUNTER2 = const(0x2)
    RESTART_COUNTER4 = const(0x4)
    RESTART_COUNTER8 = const(0x8)
    STATUS_LIGHT_NONE = const(0x0)
    STATUS_LIGHT_MONO = const(0x10)
    STATUS_LIGHT_RGB_NO_FADE = const(0x20)
    STATUS_LIGHT_RGB_FADE = const(0x30)
    SUPPORTS_ACK = const(0x100)
    SUPPORTS_BROADCAST = const(0x200)
    SUPPORTS_FRAMES = const(0x400)
    IS_CLIENT = const(0x800)
    SUPPORTS_RELIABLE_COMMANDS = const(0x1000)


JD_CONTROL_CMD_SERVICES = const(JD_CMD_ANNOUNCE)
JD_CONTROL_CMD_NOOP = const(0x80)
JD_CONTROL_CMD_IDENTIFY = const(0x81)
JD_CONTROL_CMD_RESET = const(0x82)
JD_CONTROL_CMD_FLOOD_PING = const(0x83)
JD_CONTROL_CMD_SET_STATUS_LIGHT = const(0x84)
JD_CONTROL_CMD_PROXY = const(0x85)
JD_CONTROL_CMD_RELIABLE_COMMANDS = const(0x86)
JD_CONTROL_REG_RESET_IN = const(0x80)
JD_CONTROL_REG_DEVICE_DESCRIPTION = const(0x180)
JD_CONTROL_REG_PRODUCT_IDENTIFIER = const(0x181)
JD_CONTROL_REG_BOOTLOADER_PRODUCT_IDENTIFIER = const(0x184)
JD_CONTROL_REG_FIRMWARE_VERSION = const(0x185)
JD_CONTROL_REG_MCU_TEMPERATURE = const(0x182)
JD_CONTROL_REG_UPTIME = const(0x186)
JD_CONTROL_PACK_FORMATS = {
    JD_CONTROL_CMD_FLOOD_PING: "u32 u32 u8",
    JD_CONTROL_CMD_SET_STATUS_LIGHT: "u8 u8 u8 u8",
    JD_CONTROL_CMD_RELIABLE_COMMANDS: "u32",
    JD_CONTROL_REG_RESET_IN: "u32",
    JD_CONTROL_REG_DEVICE_DESCRIPTION: "s",
    JD_CONTROL_REG_PRODUCT_IDENTIFIER: "u32",
    JD_CONTROL_REG_BOOTLOADER_PRODUCT_IDENTIFIER: "u32",
    JD_CONTROL_REG_FIRMWARE_VERSION: "s",
    JD_CONTROL_REG_MCU_TEMPERATURE: "i16",
    JD_CONTROL_REG_UPTIME: "u64"
}



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, Client
from .constants import *
from typing import Optional


class TrafficLightClient(Client):
    """
    Controls a mini traffic with a red, orange and green LED.
    Implements a client for the `Traffic Light <https://microsoft.github.io/jacdac-docs/services/trafficlight>`_ service.

    """

    def __init__(self, bus: Bus, role: str) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_TRAFFIC_LIGHT, JD_TRAFFIC_LIGHT_PACK_FORMATS, role)


    @property
    def red(self) -> Optional[bool]:
        """
        The on/off state of the red light., 
        """
        return self.register(JD_TRAFFIC_LIGHT_REG_RED).bool_value()

    @red.setter
    def red(self, value: bool) -> None:
        self.register(JD_TRAFFIC_LIGHT_REG_RED).set_values(value)


    @property
    def yellow(self) -> Optional[bool]:
        """
        The on/off state of the yellow light., 
        """
        return self.register(JD_TRAFFIC_LIGHT_REG_YELLOW).bool_value()

    @yellow.setter
    def yellow(self, value: bool) -> None:
        self.register(JD_TRAFFIC_LIGHT_REG_YELLOW).set_values(value)


    @property
    def green(self) -> Optional[bool]:
        """
        The on/off state of the green light., 
        """
        return self.register(JD_TRAFFIC_LIGHT_REG_GREEN).bool_value()

    @green.setter
    def green(self, value: bool) -> None:
        self.register(JD_TRAFFIC_LIGHT_REG_GREEN).set_values(value)


    



# Autogenerated file. Do not edit.
from jacdac.bus import Bus, SensorClient
from .constants import *
from typing import Optional, Tuple


class ColorClient(SensorClient):
    """
    Senses RGB colors
    Implements a client for the `Color <https://microsoft.github.io/jacdac-docs/services/color>`_ service.

    """

    def __init__(self, bus: Bus, role: str, *, missing_color_value: Tuple[float, float, float] = None) -> None:
        super().__init__(bus, JD_SERVICE_CLASS_COLOR, JD_COLOR_PACK_FORMATS, role)
        self.missing_color_value = missing_color_value

    @property
    def color(self) -> Optional[Tuple[float, float, float]]:
        """
        Detected color in the RGB color space., red: /,green: /,blue: /
        """
        self.refresh_reading()
        return self.register(JD_COLOR_REG_COLOR).value(self.missing_color_value)

    



# Autogenerated constants for Temperature service
from enum import IntEnum
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_TEMPERATURE = const(0x1421bac7)


class TemperatureVariant(IntEnum):
    OUTDOOR = const(0x1)
    INDOOR = const(0x2)
    BODY = const(0x3)


JD_TEMPERATURE_REG_TEMPERATURE = const(JD_REG_READING)
JD_TEMPERATURE_REG_MIN_TEMPERATURE = const(JD_REG_MIN_READING)
JD_TEMPERATURE_REG_MAX_TEMPERATURE = const(JD_REG_MAX_READING)
JD_TEMPERATURE_REG_TEMPERATURE_ERROR = const(JD_REG_READING_ERROR)
JD_TEMPERATURE_REG_VARIANT = const(JD_REG_VARIANT)
JD_TEMPERATURE_PACK_FORMATS = {
    JD_TEMPERATURE_REG_TEMPERATURE: "i22.10",
    JD_TEMPERATURE_REG_MIN_TEMPERATURE: "i22.10",
    JD_TEMPERATURE_REG_MAX_TEMPERATURE: "i22.10",
    JD_TEMPERATURE_REG_TEMPERATURE_ERROR: "u22.10",
    JD_TEMPERATURE_REG_VARIANT: "u8"
}



# Autogenerated constants for Jacscript Condition service
from jacdac.constants import *
from jacdac.system.constants import *
JD_SERVICE_CLASS_JACSCRIPT_CONDITION = const(0x1196796d)
JD_JACSCRIPT_CONDITION_CMD_SIGNAL = const(0x80)
JD_JACSCRIPT_CONDITION_EV_SIGNALLED = const(JD_EV_CHANGE)
JD_JACSCRIPT_CONDITION_PACK_FORMATS = {

}
